import numpy as np
import scipy.signal
import matplotlib.pyplot as plt
import torch
from tqdm.auto import tqdm

import time
import gc

from numba import jit, njit, prange

from pathlib import Path


def make_dFoF(
    F, 
    Fneu=None, 
    neuropil_fraction=0.7, 
    percentile_baseline=30, 
    rolling_percentile_window=None,
    roll_centered=True,
    roll_stride=1,
    roll_interpolation='linear',
    channelOffset_correction=0,
    multicore_pref=False, 
    verbose=True,
):
    """
    Calculates the dF/F and other signals. Designed for Suite2p data.
    If Fneu is left empty or =None, then no neuropil subtraction done.
    See S2p documentation for more details
    RH 2021-2023
    
    Args:
        F (np.ndarray): 
            raw fluorescence values of each ROI. shape=(ROI#, time)
        Fneu (np.ndarray): 
            Neuropil signals corresponding to each ROI. dims match F.
        neuropil_fraction (float): 
            value, 0-1, of neuropil signal (Fneu) to be subtracted off of ROI signals (F)
        percentile_baseline (float/int): 
            value, 0-100, of percentile to be subtracted off from signals
        rolling_percentile_window (int):
            window size for rolling percentile. 
            NOTE: this value will be divided by stride_roll.
            If None, then a single percentile is calculated for the entire trace.
        roll_centered (bool):
            If True, then the rolling percentile is calculated with a centered window.
            If False, then the rolling percentile is calculated with a trailing window
             where the right edge of the window is the current timepoint.
        roll_stride (int):
            Stride for rolling percentile.
            NOTE: rolling_percentile_window will be divided by this value.
            If 1, then the rolling percentile is calculated
             at every timepoint. If 2, then the rolling percentile is calculated at every
             other timepoint, etc.
        roll_interpolation (str):
            Interpolation method for rolling percentile.
            Options: 'linear', 'nearest'
            See pandas.DataFrame.rolling for more details
        channelOffset_correction (float):
            value to be added to F and Fneu to correct for channel offset
        verbose (bool): 
            Whether you'd like printed updates
    Returns:
        dFoF (np.ndarray): 
            array, dF/F
        dF (np.ndarray): 
            array
        F_neuSub (np.ndarray): 
            F with neuropil subtracted
        F_baseline (np.ndarray): 
            1-D array of size F.shape[0]. Baseline value for each ROI
    """
    from .timeSeries import percentile_numba

    tic = time.time()

    roll_stride = int(roll_stride)

    F = torch.as_tensor(F, dtype=torch.float32) + channelOffset_correction
    Fneu = torch.as_tensor(Fneu, dtype=torch.float32) + channelOffset_correction if Fneu is not None else 0

    if Fneu is None:
        F_neuSub = F
    else:
        F_neuSub = F - neuropil_fraction*Fneu

    if rolling_percentile_window is None:
        F_baseline = percentile_numba(F_neuSub.numpy(), ptile=percentile_baseline) if multicore_pref else np.percentile(F_neuSub.numpy() , percentile_baseline , axis=1)
    else:
        from .timeSeries import rolling_percentile_pd
        F_baseline = rolling_percentile_pd(
            F_neuSub.numpy()[:,::roll_stride],
            ptile=percentile_baseline, 
            window=int(rolling_percentile_window / roll_stride), 
            multiprocessing_pref=multicore_pref, 
            prog_bar=verbose,
            center=roll_centered,
            interpolation=roll_interpolation,
        )
    F_baseline = torch.as_tensor(F_baseline, dtype=torch.float32)
    F_baseline = torch.tile(F_baseline[:,:,None], (1,1,roll_stride)).reshape((F_baseline.shape[0],-1))[:,:F_neuSub.shape[1]] if roll_stride>1 else F_baseline

    dF = F_neuSub - F_baseline[:,None] if F_baseline.ndim==1 else F_neuSub - F_baseline
    dFoF = dF / F_baseline[:,None] if F_baseline.ndim==1 else dF / F_baseline

    if verbose:
        print(f'Calculated dFoF. Total elapsed time: {round(time.time() - tic,2)} seconds')
    
    return dFoF.numpy() , dF.numpy() , F_neuSub.numpy() , F_baseline.numpy()


def import_s2p(dir_s2p):
    """
    Imports suite2p data

    Args:
        dir_s2p (str):
            Path to suite2p directory

    Returns:
        F (np.ndarray):
            Fluorescence traces.
            From S2p. shape=[Neurons, Time]
        Fneu (np.ndarray):
            Fluorescence Neuropil traces.
            From S2p. shape=[Neurons, Time]
        iscell (np.ndarray):
            Boolean array indicating whether each ROI is a cell.
            From S2p. shape=[Neurons,]
        ops (dict):
            Dictionary of suite2p options.
            From S2p.
        spks (np.ndarray):
            Spike inference traces.
            From S2p. shape=[Neurons, Time]
        stat (np.ndarray):
            ROI statistics.
            From S2p. shape=[Neurons,]
    """
    dir_s2p = Path(dir_s2p).resolve()

    try:
        F = np.load(dir_s2p / 'F.npy')
    except FileNotFoundError:
        print(f'F.npy not found in {dir_s2p}')
        F = None

    try:
        Fneu = np.load(dir_s2p / 'Fneu.npy')
    except FileNotFoundError:
        print(f'Fneu.npy not found in {dir_s2p}')
        Fneu = None

    try:
        iscell = np.load(dir_s2p / 'iscell.npy')
    except FileNotFoundError:
        print(f'iscell.npy not found in {dir_s2p}')
        iscell = None

    try:
        ops = np.load(dir_s2p / 'ops.npy', allow_pickle=True)[()]
    except FileNotFoundError:
        print(f'ops.npy not found in {dir_s2p}')
        ops = None

    try:
        spks = np.load(dir_s2p / 'spks.npy')
    except FileNotFoundError:
        print(f'spks.npy not found in {dir_s2p}')
        spks = None

    try:
        stat = np.load(dir_s2p / 'stat.npy', allow_pickle=True)
    except FileNotFoundError:
        print(f'stat.npy not found in {dir_s2p}')
        stat = None
        
    return F, Fneu, iscell, ops, spks, stat


@njit(parallel=True)
def peter_noise_levels(dFoF, frame_rate):

    """"
    adapted from:  Peter Rupprecht (github: CASCADE by PTRRupprecht)
    
    Computes the noise levels for each neuron of the input matrix 'dFoF'.

    The noise level is computed as the median absolute dF/F difference
    between two subsequent time points. This is a outlier-robust measurement
    that converges to the simple standard deviation of the dF/F trace for
    uncorrelated and outlier-free dF/F traces.

    Afterwards, the value is divided by the square root of the frame rate
    in order to make it comparable across recordings with different frame rates.


    input: dFoF (matrix with n_neurons x time_points)
    output: vector of noise levels for all neurons

    """

    def abs_numba(X):
        Y = np.zeros_like(X)
        for ii in prange(len(X)):
            Y[ii] = abs(X[ii])
        return Y
    noise_levels = np.zeros(dFoF.shape[0])
    for ii in prange(dFoF.shape[0]):
        noise_levels[ii] = np.median(abs_numba(np.diff(dFoF[ii,:],1)))

    # noise_levels = np.median(np.abs(np.diff(dFoF, axis=1)), axis=1)
    noise_levels = noise_levels / np.sqrt(frame_rate) * 100    # scale noise levels to percent
    return noise_levels

def snr_autoregressive(
    x, 
    axis=1, 
    center=True, 
    standardize=True, 
    device='cpu',
    return_numpy=True,
    return_cpu=True,
):
    """
    Calculate the SNR of an autoregressive signal.
    Relies on the assumption that the magnitude of the signal
     can be estimated as the correlation of a signal and
     its autoregressive component (corr(sig, roll(sig, 1))).
    RH 2023

    Args:
        x (np.ndarray): 
            2D array of shape (n_traces, n_samples)
        axis (int, optional):
            Axis along which to calculate the SNR. Defaults to 1.
        center (bool, optional):
            Whether to center the data before calculating the SNR.
            Defaults to True.
        standardize (bool, optional):
            Whether to standardize the data before calculating the SNR.
              Defaults to True.

    Returns:
        snr (np.ndarray):
            1D array of shape (n_traces,) containing the SNR of each trace.
        s (np.ndarray):
            1D array of shape (n_traces,) containing the signal variance of each trace.
        n (np.ndarray):
            1D array of shape (n_traces,) containing the noise variance of each trace.
    """
    x_norm = torch.as_tensor(x, dtype=torch.float32, device=device)

    if center:
        x_norm = x_norm - torch.mean(x_norm, axis=axis, keepdims=True)
    if standardize:
        x_norm = x_norm / torch.std(x_norm, axis=axis, keepdims=True)

    var = ((x_norm**2).sum(axis) / (x_norm.shape[axis]-1))  ## total variance of each trace
    
    s = ((x_norm[:,1:] * x_norm[:,:-1]).sum(axis) / (x_norm.shape[axis]-2))  ## signal variance of each trace based on assumption that trace = signal + noise, signal is autoregressive, noise is not autoregressive
    n = var - s
    snr = s / n

    if return_numpy:
        snr, s, n = snr.cpu().numpy(), s.cpu().numpy(), n.cpu().numpy()
    elif return_cpu:
        snr, s, n = snr.cpu(), s.cpu(), n.cpu()

    return snr, s, n


def derivative_MAD(
    X, 
    n=(0,1,2),
    dt=1, 
    axis=1, 
    center=True, 
    standardize=False,
    device='cpu',
    return_numpy=True,
    return_cpu=True,
):
    """
    Calculate the median absolute deviance of the nth derivatives of a signal.
    This is a generalization of Peter Rupperecht's noise level calculation for
     CASCADE by PTRRupprecht (github.com/PTRRupprecht/CASCADE).
    RH 2023
    
    Args:
        X (np.ndarray or torch.Tensor): 
            Signal to calculate the MAD of the nth derivatives of.
        n (tuple):
            Tuple of integers specifying the nth derivatives to calculate.
        dt (float):
            Time step between samples of the signal. 1/frame_rate.
        axis (int):
            Axis along which to calculate the MAD of the nth derivatives.
        center (bool):
            Whether to center the signal before calculating the MAD.
            Should generally be True
        standardize (bool):
            Whether to standardize the signal before calculating the MAD.
            Should generally be False
        device (str):
            Device to use for torch tensors. 'cpu' or 'cuda'.
        return_numpy (bool):
            Whether to return the results as a numpy array.
        return_cpu (bool):
            Whether to return the results on the cpu.
    """
    n = (n,) if isinstance(n, int) else n

    ## make robust to torch tensor or numpy array inputs of X
    x = torch.as_tensor(X, dtype=torch.float32, device=device)
    
    x_norm = x - torch.mean(x, axis=axis, keepdim=True) if center else x
    x_norm = x_norm / torch.std(x, axis=axis, keepdim=True) if standardize else x_norm

    ## calculate the nth derivatives of the signal
    x_deriv = [torch.diff(x_norm, n=n_i, dim=axis) for n_i in n]

    ## calculate the median absolute deviation of the nth derivatives
    x_deriv_MAD = [torch.median(torch.abs(x_deriv_i), axis=axis)[0] / (dt**-1) for x_deriv_i in x_deriv]

    ## return as numpy array if desired
    if return_numpy:
        x_deriv_MAD = [x_deriv_MAD_i.cpu().numpy() for x_deriv_MAD_i in x_deriv_MAD]
    elif return_cpu:
        x_deriv_MAD = [x_deriv_MAD_i.cpu() for x_deriv_MAD_i in x_deriv_MAD]
    
    return x_deriv_MAD


def trace_quality_metrics(
    F,
    Fneu,
    dFoF,
    F_neuSub,
    F_baseline_roll=None,
    percentile_baseline=30,
    window_rolling_baseline=30*60*15,
    Fs=30,
    plot_pref=True, 
    thresh=None,
    device='cpu',
):
    '''
    Some simple quality metrics for calcium imaging traces. Designed to
    work with Suite2p's outputs (F, Fneu) and the make_dFoF function
    above.
    RH 2021

    Args:
        F (np.ndarray or torch.Tensor):
            Fluorescence traces. 
            From S2p. shape=[Neurons, Time]
        Fneu (np.ndarray or torch.Tensor):
            Fluorescence Neuropil traces. 
            From S2p. shape=[Neurons, Time]
        dFoF (np.ndarray or torch.Tensor):
            Normalized changes in fluorescence ('dF/F').
            From 'make_dFoF' above.
            ((F-Fneu) - F_base) / F_base . Where F_base is
            something like percentile((F-Fneu), 30)
        F_neuSub (np.ndarray or torch.Tensor):
            Neuropil subtracted fluorescence.
            From 'make_dFoF' above
        F_baseline_roll (np.ndarray or torch.Tensor):
            Rolling baseline of F_neuSub.
            From 'make_dFoF' above. If None, then
             will be calculated from F_neuSub.
        percentile_baseline (int, 0 to 100):
            percentile to use as 'baseline'
        window_rolling_baseline (int):
            Window to use for rolling baseline.
            In samples.
        Fs (float):
            Framerate of imaging
        plot_pref (bool):
            Whether to plot the traces and metrics
        thresh:
            Dictionary of thresholds to use.
            Defined as tuples of (min, max) values.
            If None, then use default values:
                'var_ratio__Fneu_over_F': (0, 0.5),
                'EV__F_by_Fneu': (0, 0.5),
                'base_FneuSub': (75, 1500),
                'base_F': (200, 2000),
                'nsr_autoregressive': (0, 6),
                'noise_derivMAD': (0, 0.02),
                'max_dFoF': (0, 10),
                'baseline_var': (0, 0.01),
        device (str):
            Device to use for torch tensors. 'cpu' or 'cuda'.
    
    Returns:
        tqm: dict with the following fields:
            metrics:
                quality_metrics. Dict of all the 
                relevant output variables
            thresh:
                Some hardcoded thresholds for aboslute 
                cutoffs
            sign:
                Whether the thresholds for exclusion in tqm_thresh
                should be positive (>) or negative (<). Multiply
                by tqm_thresh to just do everything as (>).
        good_ROIs:
            ROIs that did not meet the exclusion creteria
    '''
    from .similarity import pairwise_orthogonalization_torch

    if F_baseline_roll is None:
        from .timeSeries import rolling_percentile_pd
        F_baseline_roll = rolling_percentile_pd(
            F_neuSub.cpu().numpy()[:,::1] if isinstance(F_neuSub, torch.Tensor) else F_neuSub[:,::1],
            ptile=percentile_baseline, 
            window=int(window_rolling_baseline), 
            multiprocessing_pref=True, 
            center=True,
            interpolation='linear',
        )

    F = torch.as_tensor(F, dtype=torch.float32, device=device)
    Fneu = torch.as_tensor(Fneu, dtype=torch.float32, device=device)
    dFoF = torch.as_tensor(dFoF, dtype=torch.float32, device=device)
    F_neuSub = torch.as_tensor(F_neuSub, dtype=torch.float32, device=device)
    F_baseline_roll = torch.as_tensor(F_baseline_roll, dtype=torch.float32, device=device)

    var_F = torch.var(F, dim=1)
    var_Fneu = torch.var(Fneu, dim=1)

    var_ratio__Fneu_over_F = var_Fneu / var_F
    var_ratio__Fneu_over_F[torch.isinf(var_ratio__Fneu_over_F)] = 0

    # var_FneuSub = torch.var(F_neuSub, dim=1)
    # EV__F_by_Fneu = 1 - (var_FneuSub / var_F)
    _, EV__F_by_Fneu, _, _ = pairwise_orthogonalization_torch(
        v1=F.T,
        v2=Fneu.T,
        center=True,
    )

    # F_baseline = torch.quantile(F, percentile_baseline/100, dim=1, keepdim=True)
    # FneuSub_baseline = torch.quantile(F_neuSub, percentile_baseline/100, dim=1, keepdim=True)
    ## For some reason, torch.quantile is absurdly slow. So we'll use kthvalue instead
    k = int(np.round((F.shape[1]-1) * percentile_baseline / 100)) + 1  ## kthvalue is 1-indexed
    F_baseline = torch.kthvalue(F, k, dim=1, keepdim=True).values
    FneuSub_baseline = torch.kthvalue(F_neuSub, k, dim=1, keepdim=True).values
    
    snr_ar, _, _ = snr_autoregressive(
        x=dFoF,
        axis=1, 
        center=True, 
        standardize=True, 
        device=device,
        return_numpy=False,
        return_cpu=False,
    )
    nsr_ar = 1 / snr_ar
    noise_derivMAD = derivative_MAD(
        X=dFoF, 
        n=2,
        dt=1/Fs, 
        axis=1,
        center=True, 
        standardize=False,
        device=device,
        return_numpy=False,
        return_cpu=False,
    )[0]

    F_baseline_roll_mean = torch.mean(F_baseline_roll, dim=1, keepdim=True)
    dFbrmOverFbrm = (F_baseline_roll - F_baseline_roll_mean) / F_baseline_roll_mean
    baseline_var = torch.var(dFbrmOverFbrm, dim=1)

    max_dFoF = torch.max(dFoF, dim=1).values


    metrics = {
        'var_ratio__Fneu_over_F': var_ratio__Fneu_over_F.cpu().numpy().squeeze(),
        'EV__F_by_Fneu': EV__F_by_Fneu.cpu().numpy().squeeze(),
        'base_FneuSub': FneuSub_baseline.cpu().numpy().squeeze(),
        'base_F': F_baseline.cpu().numpy().squeeze(),
        'nsr_autoregressive': nsr_ar.cpu().numpy().squeeze(),
        'noise_derivMAD': noise_derivMAD.cpu().numpy().squeeze(),
        'max_dFoF': max_dFoF.cpu().numpy().squeeze(),
        'baseline_var': baseline_var.cpu().numpy().squeeze(),
    }
    thresh = {
        'var_ratio__Fneu_over_F': (0, 0.5),
        'EV__F_by_Fneu': (0, 0.7),
        'base_FneuSub': (100, 2000),
        'base_F': (200, 3500),
        'nsr_autoregressive': (0, 6),
        'noise_derivMAD': (0, 0.015),
        'max_dFoF': (0.75, 10),
        'baseline_var': (0, 0.015),
    } if thresh is None else thresh

    # Exclude ROIs
    good_ROIs = np.ones(dFoF.shape[0], dtype=bool)
    classifications = dict()
    for ii, met in enumerate(metrics):
        to_exclude = (metrics[met] < thresh[met][0]) + (thresh[met][1] < metrics[met]) + np.isnan(metrics[met]) # note that if NaN then excluded
        classifications[met] = np.logical_not(to_exclude)
        good_ROIs[to_exclude] = False

    # drop everything into a dict
    tqm = {
        "metrics": metrics,
        "thresh": thresh,
        "classifications": classifications,
    }

    # plot
    if plot_pref:
        fig, axs = plt.subplots(len(tqm['metrics']), figsize=(7,10))
        for ii, met in enumerate(tqm['metrics']):
            axs[ii].hist(tqm['metrics'][met][np.where(good_ROIs==1)[0]], 300, histtype='step')
            axs[ii].hist(tqm['metrics'][met][np.where(good_ROIs==0)[0]], 300, histtype='step')

            axs[ii].title.set_text(f"{met}: {np.sum(tqm['classifications'][met]==0)} excl")
            axs[ii].set_yscale('log')

            axs[ii].plot(np.array([tqm['thresh'][met],tqm['thresh'][met]])  ,  np.array([0,100]), 'k')
        fig.legend(('thresh', 'included','excluded'))
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])

        plt.figure()
        plt.plot(good_ROIs)
        plt.plot(scipy.signal.savgol_filter(good_ROIs, 101,1))

    print(f'ROIs excluded: {int(np.sum(1-good_ROIs))} / {len(good_ROIs)}')
    print(f'ROIs included: {int(np.sum(good_ROIs))} / {len(good_ROIs)}')

    good_ROIs = np.array(good_ROIs, dtype=bool)
    return tqm, good_ROIs


def get_chan1_offset(path_to_tif):
    from ScanImageTiffReader import ScanImageTiffReader

    vol = ScanImageTiffReader(path_to_tif)
    md = vol.metadata().split("\n")
    for ii, val in enumerate(md):
        if val[:25] == 'SI.hScan2D.channelOffsets':
            chan1_offset = int(val[29:33])

    print(f'chan1_offset= {chan1_offset}')
    return chan1_offset
def get_ScanImage_tiff_metadata(path_to_tif, verbose=False):
    from ScanImageTiffReader import ScanImageTiffReader
    
    vol=ScanImageTiffReader(path_to_tif)
    md = vol.metadata().split("\n")
    print(md) if verbose else None
    return md

def import_tiffs_SI(
    paths, 
    downsample_factors=[1,1,1],
    clip_zero=False,
    dtype=None,
    verbose=True,
):
    """
    Imports a tif file using ScanImageTiffReader
    RH 2023

    Args:
        paths (list):
            List of paths to tif files.
        downsample_factors (list):
            List of downsample factors for each dimension.
            [frames, height, width]
            Calls .image_processing.bin_array(frames, downsample_factors, function=partial(np.mean, axis=0))
        clip_zero (bool):
            Whether to clip negative values to zero.
        dtype (np.dtype):
            Data type to cast to.
        verbose (bool):
            Whether to show tqdm progress bar.

    Returns:
        vol (list):
            List of ScanImageTiffReader objects.
    """
    from ScanImageTiffReader import ScanImageTiffReader
    from .image_processing import bin_array
    from functools import partial

    ## warn if dtype is unsigned integer type and clip_zero is False
    if (dtype is not None) and (np.issubdtype(dtype, np.unsignedinteger)) and (not clip_zero):
        print("Warning: dtype is unsigned integer type and clip_zero is False. This will result in overflow.")

    def get_frames(path):
        vol = ScanImageTiffReader(path)
        n_frames = 1 if len(vol.shape()) < 3 else vol.shape()[0]  # single page tiffs
        frames = vol.data(beg=0, end=n_frames) if n_frames > 1 else vol.data()
        vol.close()
        del vol
        frames = np.clip(frames, 0, None) if clip_zero else frames
        frames = bin_array(frames, downsample_factors, function=partial(np.mean, axis=0)) if np.any(np.array(downsample_factors) > 1) else frames
        frames = frames.astype(dtype) if (dtype is not None) and (frames.dtype != dtype) else frames
        gc.collect()
        gc.collect()
        return frames
    
    out = [get_frames(path) for path in tqdm(paths, disable=not verbose)] if isinstance(paths, list) else get_frames(paths)

    return out

def dense_stack_to_sparse_stack_SI(
    stack_in, 
    scanimage_metadata=None,
    frames_to_discard_per_slice=30, 
    sparse_step_size_um=4,
    reduction=np.mean,
    num_frames_per_slice=60, 
    num_slices=25, 
    num_volumes=10, 
    step_size_um=0.8, 
    verbose=True,
):
    """
    Converts a dense stack of images into a sparse stack of images.
    Depends on the indexing used by ScanImage for z-stacks.
    RH 2023

    Args:
        stack_in (np.ndarray):
            Input stack of images.
            Shape: [frames, height, width]
        scanimage_metadata (list of str):
            Metadata from ScanImage tiff file. Expected to be a list of strings.
            This list can be obtained using get_ScanImage_tiff_metadata().
             fill in the following args:
                - num_frames_per_slice
                - num_slices
                - num_volumes
                - step_size_um
        frames_to_discard_per_slice (int):
            Number of frames to discard per slice.
        sparse_step_size_um (float):
            Desired step size in microns for the sparse stack.
        reduction (function):
            Function to reduce the frames in each slice.
            Must be able to accept an axis argument.
        num_frames_per_slice (int):
            Number of frames per slice.
            From SI z-stack params.
        num_slices (int):
            Number of slices.
            From SI z-stack params.
        num_volumes (int):
            Number of volumes.
            From SI z-stack params.
        step_size_um (float):
            Step size in microns.
            From SI z-stack params.

    Returns:
        stack_out (np.ndarray):
            Output stack of images.
            Shape: [frames, height, width]
        positions_z (np.ndarray):
            Z positions of each frame in the sparse stack.
            Shape: [frames]
        idx_slices (np.ndarray):
            Indices that were subsampled from the dense stack along the z-axis.
    """
    if scanimage_metadata is not None:
        meta_strings = {
            'num_frames_per_slice': 'SI.hStackManager.framesPerSlice',
            'num_slices': 'SI.hStackManager.actualNumSlices',
            'num_volumes': 'SI.hStackManager.actualNumVolumes',
            'step_size_um': 'SI.hStackManager.actualStackZStepSize',
        }
        meta_values = {key: [float(t[t.find('= ')+2:]) for t in scanimage_metadata if t[:len(m)]==m][0] for key, m in meta_strings.items()}
        num_frames_per_slice = int(meta_values['num_frames_per_slice'])
        num_slices = int(meta_values['num_slices'])
        num_volumes = int(meta_values['num_volumes'])
        step_size_um = float(meta_values['step_size_um'])

        if verbose:
            print(f"Args found from scanimage_metadata:")
            print(f"{'  num_frames_per_slice = ':<25}" + f"{num_frames_per_slice}") 
            print(f"{'  num_slices = ':<25}" + f"{num_slices}") 
            print(f"{'  num_volumes = ':<25}" + f"{num_volumes}") 
            print(f"{'  step_size_um = ':<25}" + f"{step_size_um}") 
            print('')

    range_slices = num_slices * step_size_um
    range_idx_half = int((range_slices / 2) // sparse_step_size_um)
    step_numIdx = int(sparse_step_size_um // step_size_um)
    idx_center = int(num_slices // 2)
    idx_slices = [idx_center + n for n in np.arange(-range_idx_half*step_numIdx, range_idx_half*step_numIdx + 1, step_numIdx, dtype=np.int64)]
    assert (min(idx_slices) >= 0) and (max(idx_slices) <= num_slices), f"RH ERROR: The range of slice indices expected is greater than the number of slices available: {idx_slices}"
    positions_z = [(idx - idx_center)*step_size_um for idx in idx_slices]
    positions_z = [np.round(z, decimals=10) for z in positions_z]
    
    slices_rs = np.reshape(stack_in, (num_frames_per_slice, num_slices, num_volumes, stack_in.shape[1], stack_in.shape[2]), order='F')
    slices_rs = slices_rs[frames_to_discard_per_slice:,:,:,:,:]
    slices_rs = reduction(slices_rs, axis=(0, 2))

    stack_out = slices_rs[idx_slices]

    if verbose:
        print(f"{'stack_in.shape = ':<25}" + f"{stack_in.shape}") 
        print(f"{'stack_out.shape = ':<25}" + f"{stack_out.shape}") 
        print(f"{'positions_z = ':<25}" + ''.join([f'{z}, ' for z in positions_z])) 
        print(f"{'idx_slices = ':<25}" + f"{idx_slices}")

    return stack_out, positions_z, idx_slices


def find_zShifts(
    zstack,
    positions_z=None,
    path_to_tiff=None,
    frames=None,
    clip_zero=False,
    downsample_factors=[1,1,1],
    dtype=np.uint16,
    bandpass_spatialFs_bounds=(0.02, 0.3),
    order_butter=5,
    use_GPU=True,
    batch_size=70,
    resample_factor=100,
    sig=4.0,
    verbose=True,
):
    """
    Finds the z-shift of each frame in a stack of frames relative to a z-stack.
    RH 2023

    Args:
        zstack (np.ndarray):
            Z-stack of images.
            Shape: [num_frames, height, width]
        positions_z (list):
            List of z-positions for each frame in zstack.
            Shape: zstack.shape[0]
            Can be output of dense_stack_to_sparse_stack_SI().
        path_to_tiff (str):
            Path to tiff file.
        frames (np.ndarray):
            Stack of images.
            Shape: [num_frames, height, width]
        clip_zero (bool):
            Whether to clip values below zero.
            Only used if path_to_tiff is not None.
            Used in import_tiffs_SI().
        downsample_factors (list):
            List of factors to downsample each dimension of frames by.
            Only used if path_to_tiff is not None.
            Used in import_tiffs_SI().
        dtype (np.dtype):
            Data type to convert frames to.
            Only used if path_to_tiff is not None.
            Used in import_tiffs_SI().
        bandpass_spatialFs_bounds (tuple):
            Bounds of bandpass filter in spatial frequencies.
            Used in make_Fourier_mask().
        order_butter (int):
            Order of butterworth filter.
            Used in make_Fourier_mask().
        use_GPU (bool):
            Whether to use GPU.
        batch_size (int):
            Number of frames to process at once.
            Use smaller values if GPU or CPU memory is an issue.
        resample_factor (int):
            Factor to resample frames by.
            Used to up/resample the z-axis values.
        sig (float):
            Sigma of gaussian kernel used to smooth the z-axis values.
            Used in math_functions.gaussian().
        verbose (bool):
            Whether to print progress and plot stuff.
    """
    from .image_processing import make_Fourier_mask, phaseCorrelationImage_to_shift, phase_correlation
    from .torch_helpers import clear_cuda_cache, set_device
    from .indexing import make_batches
    from .timeSeries import convolve_along_axis
    from .math_functions import gaussian

    assert positions_z is not None, "RH ERROR: Must provide positions_z"

    if frames is None:
        frames = import_tiffs_SI(
            path_to_tiff, 
            downsample_factors=downsample_factors,
            clip_zero=clip_zero,
            dtype=dtype,
            verbose=verbose,
        )
    
    mask_fft = make_Fourier_mask(
        frame_shape_y_x=frames[0].shape,
        bandpass_spatialFs_bounds=bandpass_spatialFs_bounds,
        order_butter=order_butter,
        plot_pref=verbose,
        verbose=verbose,
    )

    clear_cuda_cache() if use_GPU else None
    DEVICE = set_device(use_GPU=use_GPU, verbose=verbose)

    def frames_to_zShift(frames_toUse, zstack_maskFFT):
        def shift_helper(cc):
            return torch.stack([phaseCorrelationImage_to_shift(c)[1] for c in cc], dim=0).T
        shifts = torch.cat([shift_helper(phase_correlation(
            im_template=zstack_maskFFT,
            im_moving=batch,
            template_precomputed=True,
            device=DEVICE,
        )) for batch in make_batches(frames_toUse.astype(np.float32), batch_size=batch_size)], dim=0)
        return shifts

    zstack_maskFFT = torch.as_tensor(np.conj(np.fft.fft2(zstack, axes=(-2,-1)) * mask_fft.numpy())).to(DEVICE).type(torch.complex64)

    z_cc = frames_to_zShift(frames, zstack_maskFFT[:]).cpu().numpy()

    z_cc_conv = convolve_along_axis(z_cc[:], gaussian(x=np.linspace(-int(sig), int(sig), num=int(sig*2)+1, endpoint=True), sig=sig, plot_pref=False), axis=1, mode='same')

    xAxis = np.linspace(0, z_cc_conv.shape[1]-1, num=z_cc_conv.shape[1]*int(resample_factor), endpoint=True)
    z_cc_interp = scipy.interpolate.interp1d(np.linspace(0, z_cc_conv.shape[1]-1, num=z_cc_conv.shape[1], endpoint=True), z_cc_conv, kind='quadratic', axis=1)(xAxis)

    pos_idx_sub = np.array(positions_z[:])
    zShift_interp = scipy.interpolate.interp1d(np.linspace(0, len(pos_idx_sub)-1, num=len(pos_idx_sub), endpoint=True), np.array(positions_z[:]))(np.linspace(0, len(pos_idx_sub)-1, num=len(pos_idx_sub)*int(resample_factor), endpoint=True))

    positions_interp = zShift_interp[np.argmax(z_cc_interp, axis=1)]

    clear_cuda_cache() if use_GPU else None

    if verbose:
        plt.figure()
        plt.plot(positions_z, z_cc[0])
        plt.plot(positions_z, z_cc_conv[0])
        plt.plot(zShift_interp, z_cc_interp[0])

    return positions_interp, zShift_interp, z_cc_interp


####################################################
################ SUITE2p SPECIFIC ##################
####################################################

import pathlib
import copy

import numpy as np
import matplotlib.pyplot as plt


###############################################################################
############################## IMPORT STAT FILES ##############################
###############################################################################

def statFile_to_spatialFootprints(
    path_statFile=None, 
    statFile=None, 
    out_height_width=[36,36], 
    max_footprint_width=241, 
    plot_pref=True
    ):
    """
    Converts a stat file to a list of spatial footprint images.
    RH 2021

    Args:
        path_statFile (pathlib.Path or str):
            Path to the stat file.
            Optional: if statFile is provided, this
             argument is ignored.
        statFile (dict):
            Suite2p stat file dictionary
            Optional: if path_statFile is provided, this
             argument is ignored.
        out_height_width (list):
            [height, width] of the output spatial footprints.
        max_footprint_width (int):
            Maximum width of the spatial footprints.
        plot_pref (bool):
            If True, plots the spatial footprints.
    
    Returns:
        sf_all (list):
            List of spatial footprints images
    """
    assert out_height_width[0]%2 == 0 and out_height_width[1]%2 == 0 , "RH: 'out_height_width' must be list of 2 EVEN integers"
    assert max_footprint_width%2 != 0 , "RH: 'max_footprint_width' must be odd"
    if statFile is None:
        stat = np.load(path_statFile, allow_pickle=True)
    else:
        stat = statFile
    n_roi = stat.shape[0]
    
    # sf_big: 'spatial footprints' prior to cropping. sf is after cropping
    sf_big_width = max_footprint_width # make odd number
    sf_big_mid = sf_big_width // 2

    sf_big = np.zeros((n_roi, sf_big_width, sf_big_width))
    for ii in range(n_roi):
        sf_big[ii , stat[ii]['ypix'] - np.int16(stat[ii]['med'][0]) + sf_big_mid, stat[ii]['xpix'] - np.int16(stat[ii]['med'][1]) + sf_big_mid] = stat[ii]['lam'] # (dim0: ROI#) (dim1: y pix) (dim2: x pix)

    sf = sf_big[:,  
                sf_big_mid - out_height_width[0]//2:sf_big_mid + out_height_width[0]//2,
                sf_big_mid - out_height_width[1]//2:sf_big_mid + out_height_width[1]//2]
    if plot_pref:
        plt.figure()
        plt.imshow(np.max(sf, axis=0)**0.2)
        plt.title('spatial footprints cropped MIP^0.2')
    
    return sf

def import_multiple_stat_files(
    paths_statFiles=None, 
    dir_statFiles=None, 
    fileNames_statFiles=None, 
    out_height_width=[36,36], 
    max_footprint_width=241, 
    plot_pref=True
    ):
    """
    Imports multiple stat files.
    RH 2021 
    
    Args:
        paths_statFiles (list):
            List of paths to stat files.
            Elements can be either str or pathlib.Path.
        dir_statFiles (str):
            Directory of stat files.
            Optional: if paths_statFiles is provided, this
             argument is ignored.
        fileNames_statFiles (list):
            List of file names of stat files.
            Optional: if paths_statFiles is provided, this
             argument is ignored.
        out_height_width (list):
            [height, width] of the output spatial footprints.
        max_footprint_width (int):
            Maximum width of the spatial footprints.
        plot_pref (bool):
            If True, plots the spatial footprints.

    Returns:
        stat_all (list):
            List of stat files.
    """
    if paths_statFiles is None:
        paths_statFiles = [pathlib.Path(dir_statFiles) / fileName for fileName in fileNames_statFiles]

    sf_all_list = [statFile_to_spatialFootprints(path_statFile=path_statFile,
                                                 out_height_width=out_height_width,
                                                 max_footprint_width=max_footprint_width,
                                                 plot_pref=plot_pref)
                  for path_statFile in paths_statFiles]
    return sf_all_list

def convert_multiple_stat_files(
    statFiles_list=None, 
    statFiles_dict=None, 
    out_height_width=[36,36], 
    max_footprint_width=241, 
    print_pref=False, 
    plot_pref=False
    ):
    """
    Converts multiple stat files to spatial footprints.
    RH 2021

    Args:
        statFiles_list (list):
            List of stat files.
        out_height_width (list):
            [height, width] of the output spatial footprints.
        max_footprint_width (int):
            Maximum width of the spatial footprints.
        plot_pref (bool):
            If True, plots the spatial footprints.
    """
    if statFiles_dict is None:
        sf_all_list = [statFile_to_spatialFootprints(statFile=statFile,
                                                    out_height_width=out_height_width,
                                                    max_footprint_width=max_footprint_width,
                                                    plot_pref=plot_pref)
                    for statFile in statFiles_list]
    else:
        sf_all_list = []
        for key, stat in statFiles_dict.items():
            if print_pref:
                print(key)
            sf_all_list.append(statFile_to_spatialFootprints(statFile=stat,
                                                    out_height_width=out_height_width,
                                                    max_footprint_width=max_footprint_width,
                                                    plot_pref=plot_pref))
    return sf_all_list


def import_and_convert_to_CellReg_spatialFootprints(
    paths_statFiles, 
    frame_height=512, 
    frame_width=1024,
    dtype=np.uint8,
    ):
    """
    Imports and converts multiple stat files to spatial footprints
     suitable for CellReg.
    Output will be a list of arrays of shape (n_roi, height, width).
    RH 2022
    """

    isInt = np.issubdtype(dtype, np.integer)

    stats = [np.load(path, allow_pickle=True) for path in paths_statFiles]
    num_rois = [stat.size for stat in stats]
    sf_all_list = [np.zeros((n_roi, frame_height, frame_width), dtype) for n_roi in num_rois]
    for ii, stat in enumerate(stats):
        for jj, roi in enumerate(stat):
            if isInt:
                lam = dtype(roi['lam'] * np.iinfo(dtype).max)
            else:
                lam = roi['lam']
            sf_all_list[ii][jj, roi['ypix'], roi['xpix']] = lam
    return sf_all_list


import numpy as np
import scipy.stats
import torch

def ttest_paired_ratio(a, b):
    """
    Computes the ratio paired t-test between two arrays of data.
    This is useful for comparing ratios of data, and can be
     thought of as a geometric mean paired t-test.
    The ratio t-test takes the logarithm of the ratio of
     a vs. b and then tests the null hypothesis that the
     mean is zero.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data. Paired 1-to-1 with b.
        b (np.ndarray or torch.Tensor):
            Array of data. Paired 1-to-1 with a.
    """
    # Compute ratio
    ratio = a/b
    # Compute log of ratio
    log_ratio = np.log(ratio)
    # Compute mean of log ratio
    mean_log_ratio = np.mean(log_ratio)
    # Compute standard error of log ratio
    stderr_log_ratio = scipy.stats.sem(log_ratio)
    # Compute t-statistic
    t_stat = mean_log_ratio/stderr_log_ratio
    # Compute p-value
    p_val = scipy.stats.t.sf(np.abs(t_stat), len(a)-1)*2

    ## above is equivalent to:
    # t_stat, p_val = scipy.stats.ttest_rel(np.log(a/b), np.zeros_like(a))
    return p_val


def geometric_mean(a, axis=0, nan_policy="omit", zero_policy="propagate"):
    """
    Computes the geometric mean of an array of data.
    This is useful for computing the geometric mean of ratios.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data.
        axis (int):
            Axis along which to compute the geometric mean.
        nan_policy (str):
            Defines how to handle nan values. \n
                * "omit" - Ignore nan values.
                * "propagate" - Propagate nan values.
                * "raise" - Raise an error if nan values are present.
        zero_policy (str):
            Defines how to handle zero values. \n
                * "raise" - Raise an error if zero values are present.
                * "propagate" - Propagate zero values.
                * "nan" - Replace zero values with nan.
    """
    if isinstance(a, (np.ndarray, list)):
        mean, nanmean, isnan, exp, log, nan = np.mean, np.nanmean, np.isnan, np.exp, np.log, np.nan
    elif isinstance(a, torch.Tensor):
        mean, nanmean, isnan, exp, log, nan = torch.mean, torch.nanmean, torch.isnan, torch.exp, torch.log, torch.nan
    else:
        raise ValueError("Data must be a numpy array or a torch tensor.")
    
    if zero_policy == "raise":
        if (a == 0).any():
            raise ValueError("Data contains zero values.")
    elif zero_policy == "nan":
        a[a == 0] = nan
    elif zero_policy == "propagate":
        pass

    if nan_policy == "omit":
        mean = nanmean
    elif nan_policy == "propagate":
        mean = mean
    elif nan_policy == "raise":
        mean = mean
        if isnan(a).any():
            raise ValueError("Data contains nan values.")
        
    return exp(mean(log(a), axis=axis))


def sparsity(a, axis=0):
    """
    A normalized dispersion index. Varies from 0 (dense) to 1 (sparse).
    Computes the lifetime sparseness of an array of data.
    From Vinje & Gallant 2000; and Willmore, Mazer, & Gallant 2011.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data.
        axis (int):
            Axis along which to compute sparsity.

    Returns:
        float: Sparsity value. 0 (dense) to 1 (sparse).
    """
    return 1 - ((a.mean(axis) ** 2) / (a ** 2).mean(axis))


def error_interval(
    data, 
    axis=0,
    error="sem",
    confidence=0.95, 
    nan_policy="omit"
):
    """
    Calculate the mean and the confidence interval of the mean.

    Args:
    data (np.ndarray): 
        The input data. Can be a list or a numpy array.
    axis (int): 
        Axis along which to compute the mean and confidence interval. Default is 0.
    error (str):
        Type of error to compute. \n
            * "std" - Standard deviation.
            * "sem" - Standard error of the mean.
            * "ci" - Confidence interval. \n
    confidence (float): 
        The confidence level for the interval. Only used if error is "ci".
    nan_policy (str):
        Defines how to handle nan values. \n
            * "omit" - Ignore nan values.
            * "propagate" - Propagate nan values.
            * "raise" - Raise an error if nan values are present.

    Returns:
    tuple:
        mean (float): 
            The mean of the data along the specified axis.
        lower (float):
            The lower bound of the confidence interval.
        upper (float):
            The upper bound of the confidence interval.
    """

    assert confidence > 0 and confidence < 1, "Confidence level must be between 0 and 1."
    assert isinstance(data, np.ndarray), "Data must be a numpy array."

    ## Prepare functions for handling nan values
    if nan_policy == "omit":
        mean, std = np.nanmean, np.nanstd
    elif nan_policy == "propagate":
        mean, std = np.mean, np.std
    elif nan_policy == "raise":
        mean, std = np.mean, np.std
        if np.isnan(data).any():
            raise ValueError("Data contains nan values.")

    # Calculate the size along the specified axis
    n = data.shape[axis]

    # Compute the mean along the specified axis
    if error in ["std", "sem", "ci"]:
        m = mean(data, axis=axis)
    
        # Compute the standard deviation along the specified axis
        sd = std(data, axis=axis)
        if error in ["sem", "ci"]:
            # Compute the standard error of the mean along the specified axis
            se = sd / np.sqrt(n)
        if error == "ci":
            # Compute the margin of error
            h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
            mean, lower, upper = m, m-h, m+h
        if error == "sem":
            mean, lower, upper = m, m-se, m+se
        if error == "std":
            mean, lower, upper = m, m-sd, m+sd
    else:
        raise ValueError("Error must be one of 'std', 'sem', or 'ci'.")
    
    return mean, lower, upper


def zscore_to_pvalue(z, two_tailed=True):
    """
    Convert a z-score to a p-value.

    Args:
    z (float): 
        The z-score.
    two_tailed (bool): 
        If True, return a two-tailed p-value. If False, return a one-tailed
        p-value.

    Returns:
        float:
            The p-value.
    """
    if two_tailed:
        return 2 * scipy.stats.norm.sf(np.abs(z))
    else:
        return scipy.stats.norm.sf(np.abs(z))
    

def pvalue_to_zscore(p, two_tailed=True):
    """
    Convert a p-value to a z-score.

    Args:
    p (float): 
        The p-value.
    two_tailed (bool): 
        If True, the p-value is two-tailed. If False, the p-value is one-tailed.

    Returns:
        float:
            The z-score.
    """
    if two_tailed:
        return scipy.stats.norm.ppf(1 - p/2)
    else:
        return scipy.stats.norm.ppf(1 - p)
    

def multiTest_two_sample_independence(
    a,
    b,
) -> dict:
    """
    Perform multiple tests for two-sample independence.
    This function performs the following tests: \n
        * Two-sample t-test
            * equal variance, unequal variance
            * two-tailed, greater, less
        * Mann-Whitney U test
            * two-tailed, greater, less

    Args:
        a (np.ndarray or torch.Tensor): 
            The first samples.
        b (np.ndarray or torch.Tensor):
            The second samples.

    Returns:
        dict:
            A dictionary containing the results of the tests.
    """
    # Perform two-sample t-test
    ttest_ind = scipy.stats.ttest_ind(a, b, equal_var=True)
    ttest_ind_unequal = scipy.stats.ttest_ind(a, b, equal_var=False)
    ttest_ind_greater = scipy.stats.ttest_ind(a, b, alternative="greater")
    ttest_ind_unequal_greater = scipy.stats.ttest_ind(a, b, equal_var=False, alternative="greater")
    ttest_ind_less = scipy.stats.ttest_ind(a, b, alternative="less")
    ttest_ind_unequal_less = scipy.stats.ttest_ind(a, b, equal_var=False, alternative="less")

    # Perform Mann-Whitney U test
    mannwhitneyu = scipy.stats.mannwhitneyu(a, b, alternative="two-sided")
    mannwhitneyu_greater = scipy.stats.mannwhitneyu(a, b, alternative="greater")
    mannwhitneyu_less = scipy.stats.mannwhitneyu(a, b, alternative="less")

    return {
        "ttest_ind": ttest_ind,
        "ttest_ind_unequal": ttest_ind_unequal,
        "ttest_ind_greater": ttest_ind_greater,
        "ttest_ind_unequal_greater": ttest_ind_unequal_greater,
        "ttest_ind_less": ttest_ind_less,
        "ttest_ind_unequal_less": ttest_ind_unequal_less,
        "mannwhitneyu": mannwhitneyu,
        "mannwhitneyu_greater": mannwhitneyu_greater,
        "mannwhitneyu_less": mannwhitneyu_less,
    }


import gc
from pathlib import Path

import h5py
import numpy as np

def close_all_h5():
    '''
    Closes all h5 objects in workspace. Not tested thoroughly.
    from here: https://stackoverflow.com/questions/29863342/close-an-open-h5py-data-file
    '''
    try:
        for obj in gc.get_objects():   # Browse through ALL objects
            if isinstance(obj, h5py.File):   # Just HDF5 files
                try:
                    obj.close()
                except:
                    pass # Was already closed
    except Exception as e:
        print(f"Error closing h5 files. Will try again using `tables._open_files.close_all()`. Error: {e}")
        import tables
        tables.file._open_files.close_all()
    
    gc.collect()



def show_group_items(hObj):
    '''
    Simple function that displays all the items and groups in the
     highest hierarchical level of an h5 object or python dict.
    See 'show_item_tree' to view the whole tree
    RH 2021

    Args:
        hObj: 'hierarchical Object' hdf5 object or subgroup object OR python dictionary
    
    ##############

    example usage:
        with h5py.File(path , 'r') as f:
            h5_handling.show_group_items(f)
    '''

    for ii,val in enumerate(list(iter(hObj))):
        if isinstance(hObj[val] , h5py.Group) or isinstance(hObj[val], h5py.File):
            print(f'{ii+1}. {val}:----------------')
        if isinstance(hObj[val] , dict):
            print(f'{ii+1}. {val}:----------------')
        else:
            if hasattr(hObj[val] , 'shape') and hasattr(hObj[val] , 'dtype'):
                print(f'{ii+1}. {val}:   shape={hObj[val].shape} , dtype={hObj[val].dtype}')
            else:
                print(f'{ii+1}. {val}:   type={type(hObj[val])}')



def show_item_tree(hObj=None , path=None, depth=None, show_metadata=True, print_metadata=False, indent_level=0):
    '''
    Recursive function that displays all the items 
     and groups in an h5 object or python dict
    RH 2021

    Args:
        hObj:
            'hierarchical Object'. hdf5 object OR python dictionary
        path (Path or string):
            If not None, then path to h5 object is used instead of hObj
        depth (int):
            how many levels deep to show the tree
        show_metadata (bool): 
            whether or not to list metadata with items
        print_metadata (bool): 
            whether or not to show values of metadata items
        indent_level: 
            used internally to function. User should leave blank

    ##############
    
    example usage:
        with h5py.File(path , 'r') as f:
            h5_handling.show_item_tree(f)
    '''

    if depth is None:
        depth = int(10000000000000000000)
    else:
        depth = int(depth)

    if depth < 0:
        return

    if path is not None:
        with h5py.File(path , 'r') as f:
            show_item_tree(hObj=f, path=None, depth=depth-1, show_metadata=show_metadata, print_metadata=print_metadata, indent_level=indent_level)
    else:
        indent = f'  '*indent_level
        if hasattr(hObj, 'attrs') and show_metadata:
            for ii,val in enumerate(list(hObj.attrs.keys()) ):
                if print_metadata:
                    print(f'{indent}METADATA: {val}: {hObj.attrs[val]}')
                else:
                    print(f'{indent}METADATA: {val}: shape={hObj.attrs[val].shape} , dtype={hObj.attrs[val].dtype}')
        
        for ii,val in enumerate(list(iter(hObj))):
            if isinstance(hObj[val], h5py.Group):
                print(f'{indent}{ii+1}. {val}:----------------')
                show_item_tree(hObj[val], depth=depth-1, show_metadata=show_metadata, print_metadata=print_metadata , indent_level=indent_level+1)
            elif isinstance(hObj[val], dict):
                print(f'{indent}{ii+1}. {val}:----------------')
                show_item_tree(hObj[val], depth=depth-1, show_metadata=show_metadata, print_metadata=print_metadata , indent_level=indent_level+1)
            else:
                if hasattr(hObj[val], 'shape') and hasattr(hObj[val], 'dtype'):
                    print(f'{indent}{ii+1}. {val}:    '.ljust(20) + f'shape={hObj[val].shape} ,'.ljust(20) + f'dtype={hObj[val].dtype}')
                else:
                    print(f'{indent}{ii+1}. {val}:    '.ljust(20) + f'type={type(hObj[val])}')


def make_h5_tree(dict_obj , h5_obj , group_string='', use_compression=False, track_order=True):
    '''
    This function is meant to be called by write_dict_to_h5. It probably shouldn't be called alone.
    This function creates an h5 group and dataset tree structure based on the hierarchy and values within a python dict.
    There is a recursion in this function.
    RH 2021
    '''
    ## Set track_order to True to keep track of the order of the items in the dict
    ##  This is useful for reading the dict back in from the h5 file
    h5py.get_config().track_order = track_order
    for ii,(key,val) in enumerate(dict_obj.items()):
        if group_string=='':
            group_string='/'
        if isinstance(val , dict):
            # print(f'making group:  {key}')
            h5_obj[group_string].create_group(key)
            make_h5_tree(val , h5_obj[group_string] , f'{group_string}/{key}', use_compression=use_compression)
        else:
            ## cast to 'S' type if string so that it doesn't become '|O' object type in h5 file
            if isinstance(val, str):
                val = np.array(val, dtype=np.string_)
            
            # print(f'saving:  {group_string}: {key}')
            kwargs_compression = {'compression': 'gzip', 'compression_opts': 9} if use_compression else {}
            h5_obj[group_string].create_dataset(key , data=val, **kwargs_compression)
def write_dict_to_h5(
    path_save, 
    input_dict, 
    use_compression=False, 
    track_order=True, 
    write_mode='w-', 
    show_item_tree_pref=True
):
    '''
    Writes an h5 file that matches the hierarchy and data within a python dict.
    This function calls the function 'make_h5_tree'
    RH 2021
   
    Args:
        path_save (string or Path): 
            Full path name of file to write
        input_dict (dict): 
            Dict containing only variables that can be written as a 'dataset' in an h5 file (generally np.ndarrays and strings)
        use_compression (bool):
            Whether or not to use compression when writing the h5 file
        track_order (bool):
            Whether or not to keep track of the order of the items in the dict
        write_mode ('w' or 'w-'): 
            The priveleges of the h5 file object. 'w' will overwrite. 'w-' will not overwrite
        show_item_tree_pref (bool): 
            Whether you'd like to print the item tree or not
    '''
    with h5py.File(path_save , write_mode) as hf:
        make_h5_tree(input_dict , hf , '', use_compression=use_compression, track_order=track_order)
        if show_item_tree_pref:
            print(f'==== Successfully wrote h5 file. Displaying h5 hierarchy ====')
            show_item_tree(hf)


def simple_load(filepath, return_dict=True, verbose=False):
    """
    Returns a dictionary object containing the groups
    as keys and the datasets as values from
    given hdf file.
    RH 2023

    Args:
        filepath (string or Path): 
            Full path name of file to read.
        return_dict (bool):
            Whether or not to return a dict object (True)
            or an h5py object (False)
    """
    if return_dict:
        with h5py.File(filepath, 'r') as h5_file:
            if verbose:
                print(f'==== Loading h5 file "{filepath}" ====')
                show_item_tree(h5_file)
            result = {}
            def visitor_func(name, node):
                # Split name by '/' and reduce to nested dict
                keys = name.split('/')
                sub_dict = result
                for key in keys[:-1]:
                    sub_dict = sub_dict.setdefault(key, {})

                if isinstance(node, h5py.Dataset):
                    sub_dict[keys[-1]] = node[...]
                elif isinstance(node, h5py.Group):
                    sub_dict.setdefault(keys[-1], {})

            h5_file.visititems(visitor_func)            
            return result
    else:
        result = h5py.File(filepath, 'r')
        show_item_tree(result) if verbose else None
        return result

def h5Obj_to_dict(hObj):
    '''
    Converts an h5py object to a python dict object
    RH 2023
    '''
    h5_dict = {}
    for ii,val in enumerate(list(iter(hObj))):
        if isinstance(hObj[val], h5py.Group):
            h5_dict[val] = h5Obj_to_dict(hObj[val])
        else:
            h5_dict[val] = hObj[val][()]
    return h5_dict


def simple_save(
    dict_to_save, 
    path=None, 
    use_compression=False, 
    track_order=True, 
    write_mode='w-', 
    mkdir=True,
    verbose=False,
):
    """
    Saves a python dict to an hdf file.
    Also allows for adding new data to
     an existing hdf file.
    RH 2021

    Args:
        dict_to_save (dict):
            Python dict to save to hdf file.
        path (string or Path):
            Full path name of file to write.
        write_mode ('w', 'w-', 'a'):
            The priveleges of the h5 file object.
            'w' will overwrite.
            'w-' will not overwrite.
            'a' will append/add a new dataset to the h5 file.
        use_compression (bool):
            Whether or not to use compression when writing the h5 file
        track_order (bool):
            Whether or not to keep track of the order of the items in the dict
        verbose (bool):
            Whether or not to print out the h5 file hierarchy.
    """

    if mkdir:
        Path(path).parent.mkdir(parents=True, exist_ok=True)

    write_dict_to_h5(
        path, 
        dict_to_save, 
        use_compression=use_compression, 
        track_order=track_order,
        write_mode=write_mode, 
        show_item_tree_pref=verbose
    )


def merge_helper(d, group):
    """
    Merge a dictionary into an existing HDF5 file identified
     by an h5py.File object.

    Args:
        d (dict): 
            The dictionary containing the data to be merged.
        group (h5py.Group): 
            The HDF5 group to which the data should be merged.
    """
    for key, value in d.items():
        if isinstance(value, dict):
            # If the value is a dictionary, check if the group already exists, and either skip it or merge the data
            if key in group:
                merge_helper(value, group[key])
            else:
                subgroup = group.create_group(key)
                merge_helper(value, subgroup)
        else:
            # If the value is not a dictionary, convert it to a numpy array and create a dataset
            if key in group:
                del group[key]
            group.create_dataset(key, data=value)
def merge_dict_into_h5_file(d, filepath=None, h5Obj=None,):
    """
    Merge a dictionary into an existing HDF5 file identified
     by a file path.
    This function wraps a recursive function that goes through
     each hierarchical level of the input dictionary and merges
     it into the appropriate HDF5 group.

    Args:
        d (dict): 
            The dictionary containing the data to be merged.
        filepath (str): 
            The file path of the HDF5 file.
            Do not specify if hObj is specified.
        h5Obj (h5py.File):
            An h5py.File object.
            Do not specify if filepath is specified.
    """
    if filepath is None and h5Obj is None:
        raise ValueError('Either filepath or h5Obj must be specified.')
    elif filepath is not None and h5Obj is not None:
        raise ValueError('Only one of filepath or h5Obj must be specified.')
    
    elif filepath is not None:
        with h5py.File(filepath, 'a') as file:
            merge_helper(d, file)
    elif h5Obj is not None:
        merge_helper(d, h5Obj)



from typing import Any, Dict, List, Optional, Tuple, Union
import warnings

import numpy as np

class Equivalence_checker():
    """
    Class for checking if all items are equivalent or allclose (almost equal) in
    two complex data structures. Can check nested lists, dicts, and other data
    structures. Can also optionally assert (raise errors) if all items are not
    equivalent. 
    RH 2023

    Attributes:
        _kwargs_allclose (Optional[dict]): 
            Keyword arguments for the `numpy.allclose` function.
        _assert_mode (bool):
            Whether to raise an assertion error if items are not close.

    Args:
        kwargs_allclose (Optional[dict]): 
            Keyword arguments for the `numpy.allclose` function. (Default is
            ``{'rtol': 1e-7, 'equal_nan': True}``)
        assert_mode (bool): 
            Whether to raise an assertion error if items are not close.
        verbose (bool):
            How much information to print out:
                * ``False`` / ``0``: No information printed out.
                * ``True`` / ``1``: Mismatched items only.
                * ``2``: All items printed out.
    """
    def __init__(
        self,
        kwargs_allclose: Optional[dict] = {'rtol': 1e-7, 'equal_nan': True},
        assert_mode=False,
        verbose=False,
    ) -> None:
        """
        Initializes the Allclose_checker.
        """
        self._kwargs_allclose = kwargs_allclose
        self._assert_mode = assert_mode
        self._verbose = verbose
        
    def _checker(
        self, 
        test: Any,
        true: Any, 
        path: Optional[List[str]] = None,
    ) -> bool:
        """
        Compares the test and true values using numpy's allclose function.

        Args:
            test (Union[dict, list, tuple, set, np.ndarray, int, float, complex,
            str, bool, None]): 
                Test value to compare.
            true (Union[dict, list, tuple, set, np.ndarray, int, float, complex,
            str, bool, None]): 
                True value to compare.
            path (Optional[List[str]]): 
                The path of the data structure that is currently being compared.
                (Default is ``None``)

        Returns:
            (bool): 
                result (bool): 
                    Returns True if all elements in test and true are close.
                    Otherwise, returns False.
        """
        try:
            ## If the dtype is a kind of string (or byte string) or object, then allclose will raise an error. In this case, just check if the values are equal.
            if np.issubdtype(test.dtype, np.str_) or np.issubdtype(test.dtype, np.bytes_) or test.dtype == np.object_:
                out = bool(np.all(test == true))
            else:
                out = np.allclose(test, true, **self._kwargs_allclose)
        except Exception as e:
            out = None  ## This is not False because sometimes allclose will raise an error if the arrays have a weird dtype among other reasons.
            warnings.warn(f"WARNING. Equivalence check failed. Path: {path}. Error: {e}") if self._verbose else None
            
        if out == False:
            if self._assert_mode:
                raise AssertionError(f"Equivalence check failed. Path: {path}.")
            if self._verbose:
                ## Come up with a way to describe the difference between the two values. Something like the following:
                ### IF the arrays are numeric, then calculate the relative difference
                dtypes_numeric = (np.number, np.bool_, np.integer, np.floating, np.complexfloating)
                if any([np.issubdtype(test.dtype, dtype) and np.issubdtype(true.dtype, dtype) for dtype in dtypes_numeric]):
                    diff = np.abs(test - true)
                    at = np.abs(true)
                    r_diff = diff / at if np.all(at != 0) else np.inf
                    r_diff_mean, r_diff_max, any_nan = np.nanmean(r_diff), np.nanmax(r_diff), np.any(np.isnan(r_diff))
                    ## fraction of mismatches
                    n_elements = np.prod(test.shape)
                    n_mismatches = np.sum(diff > 0)
                    frac_mismatches = n_mismatches / n_elements
                    ## Use scientific notation and round to 3 decimal places
                    reason = f"Equivalence: Relative difference: mean={r_diff_mean:.3e}, max={r_diff_max:.3e}, any_nan={any_nan}, n_elements={n_elements}, n_mismatches={n_mismatches}, frac_mismatches={frac_mismatches:.3e}"
                else:
                    reason = f"Values are not numpy numeric types. types: {test.dtype}, {true.dtype}"
        else:
            reason = "equivlance"

        return out, reason

    def __call__(
        self,
        test: Union[dict, list, tuple, set, np.ndarray, int, float, complex, str, bool, None], 
        true: Union[dict, list, tuple, set, np.ndarray, int, float, complex, str, bool, None], 
        path: Optional[List[str]] = None,
    ) -> Dict[str, Tuple[bool, str]]:
        """
        Compares the test and true values and returns the comparison result.
        Handles various data types including dictionaries, iterables,
        np.ndarray, scalars, strings, numbers, bool, and None.

        Args:
            test (Union[dict, list, tuple, set, np.ndarray, int, float, complex,
            str, bool, None]): 
                Test value to compare.
            true (Union[dict, list, tuple, set, np.ndarray, int, float, complex,
            str, bool, None]): 
                True value to compare.
            path (Optional[List[str]]): 
                The path of the data structure that is currently being compared.
                (Default is ``None``)

        Returns:
            Dict[Tuple[bool, str]]: 
                result Dict[Tuple[bool, str]]: 
                    The comparison result as a dictionary or a tuple depending
                    on the data types of test and true.
        """
        if path is None:
            path = ['']

        if len(path) > 0:
            if path[-1].startswith('_'):
                return (None, 'excluded from testing')

        ## NP.NDARRAY
        if isinstance(true, np.ndarray):
            result = self._checker(test, true, path)
        ## NP.SCALAR
        elif np.isscalar(true):
            if isinstance(test, (int, float, complex, np.number)):
                result = self._checker(np.array(test), np.array(true), path)
            else:
                result = (test == true, 'equivalence')
        ## NUMBER
        elif isinstance(true, (int, float, complex)):
            result = self._checker(test, true, path)
        ## DICT
        elif isinstance(true, dict):
            result = {}
            for key in true:
                if key not in test:
                    result[str(key)] = (False, 'key not found')
                else:
                    result[str(key)] = self.__call__(test[key], true[key], path=path + [str(key)])
        ## ITERATABLE
        elif isinstance(true, (list, tuple, set)):
            if len(true) != len(test):
                result = (False, 'length_mismatch')
            else:
                if all([isinstance(i, (int, float, complex, np.number)) for i in true]):
                    result = self._checker(np.array(test), np.array(true), path)
                else:
                    result = {}
                    for idx, (i, j) in enumerate(zip(test, true)):
                        result[str(idx)] = self.__call__(i, j, path=path + [str(idx)])
        ## STRING
        elif isinstance(true, str):
            result = (test == true, 'equivalence')
        ## BOOL
        elif isinstance(true, bool):
            result = (test == true, 'equivalence')
        ## NONE
        elif true is None:
            result = (test is None, 'equivalence')

        ## OBJECT with __dict__
        elif hasattr(true, '__dict__'):
            result = {}
            for key in true.__dict__:
                if key.startswith('_'):
                    continue
                if not hasattr(test, key):
                    result[str(key)] = (False, 'key not found')
                else:
                    result[str(key)] = self.__call__(getattr(test, key), getattr(true, key), path=path + [str(key)])
        ## N/A
        else:
            result = (None, 'not tested')

        if isinstance(result, tuple):
            if self._assert_mode:
                assert (result[0] != False), f"Equivalence check failed. Path: {path}. Reason: {result[1]}"

            if self._verbose > 0:
                ## Print False results
                if result[0] == False:
                    print(f"Equivalence check failed. Path: {path}. Reason: {result[1]}")
            if self._verbose > 1:
                ## Print True results
                if result[0] == True:
                    print(f"Equivalence check passed. Path: {path}. Reason: {result[1]}")
                elif result[0] is None:
                    print(f"Equivalence check not tested. Path: {path}. Reason: {result[1]}")

        return result


from typing import Tuple, Optional, Union
import math
import gc
import copy

import sklearn.decomposition
import numpy as np
import scipy.interpolate
import matplotlib.pyplot as plt
import torch
from torch.utils.data import Dataset, DataLoader
from tqdm.notebook import tqdm
import tensorly as tl

# import cuml
# import cuml.decomposition
# import cupy

from . import similarity, torch_helpers


###########################
########## PCA ############
###########################


def svd_flip(
    u: torch.Tensor, 
    v: torch.Tensor,
    u_based_decision: bool = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Sign correction to ensure deterministic output from SVD.

    The output from SVD does not have a unique sign. This function corrects the
    sign of the output to ensure deterministic output from the SVD function.

    RH 2024

    Args:
        u (torch.Tensor):
            The left singular vectors.
        v (torch.Tensor):
            The right singular vectors.
        u_based_decision (bool):
            If ``True``, the sign correction is based on the left singular
            vectors. If ``False``, the sign correction is based on the right
            singular vectors.

    Returns:
        (Tuple[torch.Tensor, torch.Tensor]):
            u (torch.Tensor):
                The corrected left singular vectors.
            v (torch.Tensor):
                The corrected right singular vectors.
    """
    as_tensor = lambda x: torch.as_tensor(x) if isinstance(x, np.ndarray) else x
    u, v = (as_tensor(var) for var in (u, v))
    
    if u_based_decision:
        # For each column j in u, pick row idx_i = argmax |u[i,j]|
        # signs[j] = sign(u[idx_i, j])
        # then u[:,j] *= signs[j];  v[j,:] *= signs[j]
        max_abs_rows = torch.argmax(u.abs(), dim=0)          # shape (n_components,)
        signs = torch.sign(u[max_abs_rows, torch.arange(u.shape[1], device=u.device)])
        u = u * signs.unsqueeze(0)
        v = v * signs.unsqueeze(1)
    else:
        # For each row i in v, pick col idx_j = argmax |v[i,j]|
        # signs[i] = sign(v[i, idx_j])
        # then v[i,:] *= signs[i];  u[:,i] *= signs[i]
        max_abs_cols = torch.argmax(v.abs(), dim=1)          # shape (n_components,)
        signs = torch.sign(v[torch.arange(v.shape[0], device=v.device), max_abs_cols])
        v = v * signs.unsqueeze(1)
        if u is not None:
            u = u * signs.unsqueeze(0)
    return u, v

class PCA(torch.nn.Module, sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
    """
    Principal Component Analysis (PCA) module.

    This module performs PCA on the input data and returns the principal
    components and the explained variance. The PCA is performed using the
    singular value decomposition (SVD) method. This class follows sklearn's PCA
    implementation and style and is subclassed from torch.nn.Module,
    sklearn.base.BaseEstimator, and sklearn.base.TransformerMixin. The
    decomposed variables (components, explained_variance, etc.) are stored as
    buffers so that they are stored in the state_dict, respond to .to() and
    .cuda(), and are saved when the model is saved.

    RH 2024

    Args:
        n_components (Optional[int]):
            Number of principal components to retain. If ``None``, all
            components are retained.
        center (bool):
            If ``True``, the data is mean-subtracted before performing SVD.
        zscale (bool):
            If ``True``, the data is z-scored before performing SVD. Equivalent
            of doing eigenvalue decomposition on the correlation matrix.
        whiten (bool):
            If ``True``, the principal components are divided by the square root
            of the explained variance.
        use_lowRank (bool):
            If ``True``, the low-rank SVD is used. This is faster but less
            accurate. Uses torch.svd_lowrank instead of torch.linalg.svd.
        lowRank_niter (int):
            Number of subspace iterations for low-rank SVD. See
            torch.svd_lowrank for more details.

    Attributes:
        n_components (int):
            Number of principal components to retain.
        whiten (bool):
            If ``True``, the principal components are divided by the square root
            of the explained variance.
        device (str):
            The device where the tensors will be stored.
        dtype (torch.dtype):
            The data type to use for the tensor.
        components (torch.Tensor):
            The principal components.
        explained_variance (torch.Tensor):
            The explained variance.
        explained_variance_ratio (torch.Tensor):
            The explained variance ratio.

    Example:
        .. highlight:: python
        .. code-block:: python

            X = torch.randn(100, 10)
            pca = PCA(n_components=5)
            pca.fit(X)
            X_pca = pca.transform(X)
    """
    def __init__(
        self,
        n_components: Optional[int] = None,
        center: bool = True,
        zscale: bool = False,
        whiten: bool = False,
        use_lowRank: bool = False,
        lowRank_niter: int = 2,
    ):
        """
        Initializes the PCA module with the provided parameters.
        """
        super(PCA, self).__init__()
        self.n_components = n_components
        self.center = center
        self.zscale = zscale
        self.whiten = whiten
        self.use_lowRank = use_lowRank
        self.lowRank_niter = lowRank_niter

    def prepare_input(
        self,
        X: torch.Tensor,
        center: bool,
        zscale: bool
    ) -> torch.Tensor:
        """
        Prepares the input data for PCA.

        Args:
            X (torch.Tensor):
                The input data to prepare.
            center (bool):
                If ``True``, the data is mean-subtracted.
            zscale (bool):
                If ``True``, the data is z-scored.

        Returns:
            (torch.Tensor):
                The prepared input data.
        """
        if isinstance(X, np.ndarray):
            X = torch.as_tensor(X)
        assert isinstance(X, torch.Tensor), 'Input must be a torch.Tensor.'
        X = X[:, None] if X.ndim == 1 else X
        assert X.ndim == 2, 'Input must be 2D.'

        if center:
            mean_ = torch.mean(X, dim=0)
            X = X - mean_
            self.register_buffer('mean_', mean_)
        if zscale:
            std_ = torch.std(X, dim=0)
            X = X / std_
            self.register_buffer('std_', std_)
        return X

    def fit(
        self,
        X: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Fits the PCA module to the input data.

        Args:
            X (torch.Tensor):
                The input data to fit the PCA module to. Should be shape
                (n_samples, n_features).

        Returns:
            self (PCA object):
                Returns the PCA object.
        """
        self._fit(X)
        return self

    def _fit(
        self,
        X: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Fits the PCA module to the input data.

        Args:
            X (torch.Tensor):
                The input data to fit the PCA module to. Should be shape
                (n_samples, n_features).

        Returns:
            (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
                U (torch.Tensor):
                    The left singular vectors. Shape (n_samples, n_components).
                S (torch.Tensor):
                    The singular values. Shape (n_components,).
                V (torch.Tensor):
                    The right singular vectors. Shape (n_features,
                    n_components).
        """
        self.n_samples_, self.n_features_ = X.shape
        self.n_components_ = min(self.n_components, self.n_features_) if self.n_components is not None else self.n_features_

        X = self.prepare_input(X, center=self.center, zscale=self.zscale)
        if self.use_lowRank:
            U, S, Vh = torch.svd_lowrank(X, q=self.n_components_, niter=self.lowRank_niter)
            Vh = Vh.T  ## torch.svd_lowrank returns Vh transposed.
        else:
            U, S, Vh = torch.linalg.svd(X, full_matrices=False)  ## U: (n_samples, n_features), S: (n_features,), Vh: (n_features, n_features). Vh is already transposed.
        U, Vh = svd_flip(U, Vh, u_based_decision=False)

        explained_variance_ = S**2 / (self.n_samples_ - 1)
        explained_variance_ratio_ = explained_variance_ / torch.sum(explained_variance_)

        components_ = Vh[:self.n_components_]
        singular_values_ = S[:self.n_components_]
        explained_variance_ = explained_variance_[:self.n_components_]
        explained_variance_ratio_ = explained_variance_ratio_[:self.n_components_]

        [self.register_buffer(name, value) for name, value in zip(
            ['components_', 'singular_values_', 'explained_variance_', 'explained_variance_ratio_'],
            [components_, singular_values_, explained_variance_, explained_variance_ratio_]
        )]

        return U, S, Vh
    
    def transform(
        self,
        X: torch.Tensor,
        y: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Transforms the input data using the fitted PCA module.

        Args:
            X (torch.Tensor):
                The input data to transform.
            y (Optional[torch.Tensor]):
                Ignored. This parameter exists to match the sklearn API.

        Returns:
            (torch.Tensor):
                The transformed data. Will be shape (n_samples, n_components).
        """
        assert hasattr(self, 'components_'), 'PCA module must be fitted before transforming data.'
        X = self.prepare_input(X, center=self.center, zscale=self.zscale)
        X_transformed = X @ self.components_.T
        if self.whiten:
            X_transformed /= torch.sqrt(self.explained_variance_)
        return X_transformed
    
    def fit_transform(
        self,
        X: torch.Tensor
    ) -> torch.Tensor:
        """
        Fits the PCA module to the input data and transforms the input data.

        Args:
            X (torch.Tensor):
                The input data to fit the PCA module to and transform.

        Returns:
            (torch.Tensor):
                The transformed data.
        """
        self.n_samples_, self.n_features_ = X.shape

        U, S, V = self._fit(X)
        U = U[:, :self.n_components_]
        
        if self.whiten:
            U *= math.sqrt(self.n_samples_ - 1)
        else:
            U *= S[:self.n_components_]

        return U
    
    def inverse_transform(
        self,
        X: torch.Tensor
    ) -> torch.Tensor:
        """
        Inverse transforms the input data using the fitted PCA module.

        Args:
            X (torch.Tensor):
                The input data to inverse transform. Should be shape (n_samples,
                n_components).

        Returns:
            (torch.Tensor):
                The inverse transformed data. Will be shape (n_samples,
                n_features).
        """
        assert hasattr(self, 'components_'), 'PCA module must be fitted before transforming data.'
        X = self.prepare_input(X, center=False, zscale=False)
        
        if self.whiten:
            scaled_components = torch.sqrt(self.explained_variance_)[:, None] * self.components_
        else:
            scaled_components = self.components_
        
        X = X @ scaled_components

        if self.zscale:
            assert hasattr(self, 'std_'), 'self.zscale is True, but std_ is not found.'
            X = X * self.std_
        if self.center:
            assert hasattr(self, 'mean_'), 'self.center is True, but mean_ is not found.'
            X = X + self.mean_

        return X


def torch_pca(
    X_in, 
    device='cpu', 
    mean_sub=True, 
    zscore=False, 
    rank=None, 
    return_cpu=True, 
    return_numpy=False,
    cuda_empty_cache=True,
):
    """
    Principal Components Analysis for PyTorch.
    If using GPU, then call torch.cuda.empty_cache() after.
    RH 2021

    Args:
        X_in (torch.Tensor or np.ndarray):
            Data to be decomposed.
            2-D array. Columns are features, rows are samples.
            PCA will be performed column-wise.
        device (str):
            Device to use. ie 'cuda' or 'cpu'. Use a function 
             torch_helpers.set_device() to get.
        mean_sub (bool):
            Whether or not to mean subtract ('center') the 
             columns.
        zscore (bool):
            Whether or not to z-score the columns. This is 
             equivalent to doing PCA on the correlation-matrix.
        rank (int):
            Maximum estimated rank of decomposition. If None,
             then rank is X.shape[1]
        return_cpu (bool):  
            Whether or not to force returns/outputs to be on 
             the 'cpu' device. If False, and device!='cpu',
             then returns will be on device.
        return_numpy (bool):
            Whether or not to force returns/outputs to be
             numpy.ndarray type.
        cuda_empty_cache (bool):
            Whether or not to call torch.cuda.empty_cache()
            Only relevant if device!='cpu'.

    Returns:
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
        scores (torch.Tensor or np.ndarray):
            The scores of the decomposition.
            2-D array.
            Each column is a score vector. Each row is a 
             sample weight.
        singVals (torch.Tensor or np.ndarray):
            The singular values of the decomposition.
            1-D array.
            Each element is a singular value.
        EVR (torch.Tensor or np.ndarray):
            The explained variance ratio of each component.
            1-D array.
            Each element is the explained variance ratio of
             the corresponding component.
    """
    
    if isinstance(X_in, torch.Tensor) == False:
        X = torch.from_numpy(X_in).to(device)
    elif str(X_in.device) != device:
        X = X_in.to(device)
    else:
        X = X_in
        
    if mean_sub and not zscore:
        X = X - torch.mean(X, dim=0)
    if zscore:
        X = X - torch.mean(X, dim=0)
        stds = torch.std(X, dim=0)
        idx_zeroStd = torch.where(stds==0)[0]
        stds[idx_zeroStd] = 1.0
        X = X / stds
        
    if rank is None:
        rank = min(list(X.shape))
    
    (U,S,V) = torch.pca_lowrank(X, q=rank, center=False, niter=2)
    components = V
    scores = torch.matmul(X, V[:, :rank])

    singVals = (S**2)/(len(S)-1)
    EVR = (singVals) / torch.sum(singVals)
    
    if return_cpu:
        components = components.cpu()
        scores = scores.cpu()
        singVals = singVals.cpu()
        EVR = EVR.cpu()
    if return_numpy:
        components = components.cpu().numpy()
        scores = scores.cpu().numpy()
        singVals = singVals.cpu().numpy()
        EVR = EVR.cpu().numpy()

    if cuda_empty_cache and device!='cpu':
        torch.cuda.empty_cache()
        gc.collect()
        
    return components, scores, singVals, EVR

def unmix_pcs(pca_components, weight_vecs):
    """
    Transforms weight_vecs into pca_components space
    RH 2021
    """
    if weight_vecs.ndim == 1:
        weight_vecs = weight_vecs[:,None]
    if type(pca_components) == np.ndarray:
        zeros = np.zeros
    elif type(pca_components) == torch.Tensor:
        zeros = torch.zeros
    
    mixing_vecs = zeros((pca_components.shape[1], weight_vecs.shape[1]))
    mixing_vecs[:weight_vecs.shape[0],:] = weight_vecs

    return pca_components @ mixing_vecs


def dimensionality_pca(
    X, 
    ev=[0.10,0.25,0.50,0.90,0.95], 
    mean_sub=True, 
    zscore=False, 
    device='cpu', 
    kwargs_torchPCA={}
):
    """
    Returns the number of components needed to explain a certain 
     amount of variance. 
    Calculates the 'linear embedding dimensionality'.
    Uses torch_pca() function to perform PCA.
    RH 2022

    Args:
        X (np.ndarray or torch.Tensor):
            Data to be decomposed.
            2-D array. Columns are features, rows are samples.
        ev (list or np.ndarray):
            List of explained variance ratios to calculate the
             number of components needed to explain.
            ie [0.10,0.25,0.50,0.90,0.95]
        mean_sub (bool):
            Whether or not to mean subtract ('center') the
             columns.
        zscore (bool):
            Whether or not to z-score the columns. This is
             equivalent to doing PCA on the correlation-matrix.
        device (str):
            Device to use. ie 'cuda' or 'cpu'. Use a function
             torch_helpers.set_device() to get.

    Returns:
        n_components (np.ndarray):
            Number of components needed to explain the
             corresponding ev.
            Same array type (torch.Tensor or np.array) as X.
    """
    ## make sure ev is a list or np.ndarray
    ev = [ev] if isinstance(ev, (int,float)) else ev
    assert isinstance(ev, (list,np.ndarray)), 'RH ERROR: ev must be a list or np.ndarray.'
    ev = np.array(ev)

    ## assert there are no nans
    assert np.any(np.isnan(X)) == False, 'RH ERROR: X must not contain NaNs.'

    comps, scores, sv, EVR = torch_pca(
        X_in=X, 
        device=device,
        return_numpy=True,
        mean_sub=mean_sub,
        zscore=zscore,
        **kwargs_torchPCA,
    )
    EVR_cumsum = np.cumsum(np.concatenate([[0], EVR]))
    interp = scipy.interpolate.interp1d(x=EVR_cumsum, y=np.arange(0, len(EVR_cumsum)), kind='linear')
    return interp(ev)


def ZCA_whiten(
    X: Union[np.ndarray, torch.Tensor],
    V: Union[np.ndarray, torch.Tensor],
    S: Union[np.ndarray, torch.Tensor],
    eps: float = 1e-5,
):
    """
    ZCA whitening of data.
    See: https://jermwatt.github.io/control-notes/posts/zca_sphereing/ZCA_Sphereing.html
    RH 2024

    Args:
        X (np.ndarray or torch.Tensor):
            Data to be whitened. \n
            Shape (n_samples, n_features).
        V (np.ndarray or torch.Tensor):
            The principal components / eigenvectors. \n
            You can use PCA.components_ from sklearn.decomposition.PCA or
            PCA.components_ above. \n
            Shape (n_features, n_components).
        S (np.ndarray or torch.Tensor):
            The singular values / eigenvalues. \n
            You can use PCA.singular_values_ from sklearn.decomposition.PCA or
            PCA.singular_values_ above. \n
            Shape (n_components,).
        eps (float):
            Small value to prevent division by zero.

    Returns:
        X_zca (np.ndarray or torch.Tensor):
            The ZCA whitened data. \n
            Shape (n_samples, n_features).

    Demo:
        ..code-block:: python
        
            X = np.random.randn(100, 10)
            pca = PCA(n_components=5)
            pca.fit(X)
            X_zca = ZCA_whiten(
                X=X,
                V=pca.components_,
                S=pca.singular_values_,
                eps=1e-5,
            )
    """
    if isinstance(X, np.ndarray):
        mean, sqrt, diag = np.mean, np.sqrt, np.diag
    elif isinstance(X, torch.Tensor):
        mean, sqrt, diag = torch.mean, torch.sqrt, torch.diag

    X = X - mean(X, axis=0, keepdims=True)

    D_inv = diag(1.0 / (S + eps))
    W_zca = V.T @ D_inv @ V
    X_zca = X @ W_zca

    return X_zca


class CP_NN_HALS_minibatch:
    """
    Minibatch version of Tensorly's CP_NN_HALS. Optimization proceeds by
    randomly sampling indices along the defined batching dimension.

    RH 2024

    Args:
        kwargs_CP_NN_HALS (dict):
            Keyword arguments to pass to Tensorly's CP_NN_HALS.
        batch_dimension (int):
            Dimension along which to batch the data (0 to n_dims-1).
        batch_size (int):
            Number of samples to batch.
        n_iter_batch (int):
            Number of optimization iterations to perform on each batch.
        n_epochs (int):
            Number of epochs to perform.
        device (str):
            Device to transfer each batch to and where the factors will be
            stored.
        kwargs_dataloader (dict):
            Keyword arguments to pass to torch.utils.data.DataLoader.
        jitter_zeros (bool):
            Whether or not to jitter any 0s in the factors to be small values.
        random_state (int):
            Random seed for reproducibility.
        verbose (bool):
            Whether or not to print progress.
    """
    def __init__(
        self,
        kwargs_CP_NN_HALS: dict = {
            'rank': 10,
            'init': 'svd',
            'svd': 'truncated_svd',
            'non_negative': True,
            'random_state': 0,
            'normalize_factors': False,
            'mask': None,
            'svd_mask_repeats': 5,
        },
        batch_dimension: int = 0,
        batch_size: int = 10,
        n_iter_batch: int = 10,
        n_epochs: int = 10,
        device: str = 'cpu',
        kwargs_dataloader: dict = {
            'drop_last': True,
            'shuffle': True,
            # 'num_workers': 0,
            # 'pin_memory': False,
            # 'prefetch_factor': 2,
            # 'persistent_workers': False,
        },
        jitter_zeros: bool = False,
        random_state: int = 0,
        verbose: bool = True,
    ):
        """
        Initializes the CP_NN_HALS_minibatch module with the provided parameters.
        """
        self.kwargs_CP_NN_HALS = kwargs_CP_NN_HALS
        self.batch_dimension = int(batch_dimension)
        self.batch_size = int(batch_size)
        self.n_iter_batch = int(n_iter_batch)
        self.n_epochs = int(n_epochs)
        self.device = torch.device(device)
        self.kwargs_dataloader = dict(kwargs_dataloader)
        self.jitter_zeros = jitter_zeros
        self.random_state = random_state
        self.verbose = verbose

        self.factors = None
        self.losses = {}
        tl.set_backend('pytorch')

    def init_factors(self, X):
        """
        Initializes the factors.
        """
        kwargs_default = {
            'rank': 10,
            'init': 'svd',
            'svd': 'truncated_svd',
            'non_negative': True,
            'random_state': 0,
            'normalize_factors': False,
            'mask': None,
            'svd_mask_repeats': 5,
        }
        self.factors = tl.decomposition._cp.initialize_cp(
            tensor=X.moveaxis(self.batch_dimension, 0),
            **{k: self.kwargs_CP_NN_HALS.get(k, v) for k, v in kwargs_default.items()},
        )
        self.factors.factors = [torch.as_tensor(f).to(self.device) for f in self.factors.factors]
        self.factors.weights = torch.as_tensor(self.factors.weights).to(self.device)
        return self.factors

    def fit(self, X):
        """
        Fits the CP_NN_HALS_minibatch module to the input data.

        Args:
            X (torch.Tensor):
                The input data to fit the CP_NN_HALS_minibatch module to.

        Returns:
            self (CP_NN_HALS_minibatch object):
                Returns the CP_NN_HALS_minibatch object.
        """
        ## Clear CUDA cache
        factors_tmp = copy.deepcopy(self.factors) if self.factors is not None else None
        self.factors = None
        torch_helpers.clear_cuda_cache()
        self.factors = copy.deepcopy(factors_tmp)
        del factors_tmp

        ## Initialize factors
        self.init_factors(X) if self.factors is None else None

        ## Make dataset (if X is not a dataset)
        if isinstance(X, torch.Tensor):
            dataset = torch.utils.data.TensorDataset(
                X.moveaxis(self.batch_dimension, 0),
                torch.arange(X.shape[self.batch_dimension]),
            )
        elif isinstance(X, Dataset):
            class DatasetWrapper(Dataset):
                def __init__(self, dataset):
                    super(DatasetWrapper, self).__init__()
                    self.dataset = dataset
                def __len__(self):
                    return len(self.dataset)
                def __getitem__(self, idx):
                    return self.dataset[idx], idx
            dataset = DatasetWrapper(X)
        else:
            raise ValueError('X must be a torch.Tensor or torch.utils.data.Dataset.')
        ## Make dataloader
        kwargs_dataloader_tmp = copy.deepcopy(self.kwargs_dataloader)
        kwargs_dataloader_tmp.pop('batch_size', None)
        dataloader = torch.utils.data.DataLoader(
            dataset, 
            batch_size=self.batch_size, 
            **kwargs_dataloader_tmp,
        )

        kwargs_tmp = copy.deepcopy(self.kwargs_CP_NN_HALS)
        kwargs_tmp.pop('init', None)
        kwargs_tmp.pop('n_iter_max', None)
        ## Fit
        for i_epoch in tqdm(range(self.n_epochs), desc='CP_NN_HALS_minibatch', leave=True, disable=self.verbose==False):
            for i_iter, (X_batch, idx_batch) in tqdm(enumerate(dataloader), desc='Batch', leave=False, disable=self.verbose==False, total=len(dataloader)):
                X_batch = X_batch.to(self.device)
                idx_batch = idx_batch.to(self.device)
                
                factors_batch = copy.deepcopy(self.factors)
                factors_batch.factors = [torch.as_tensor(f, device=self.device) for f in factors_batch.factors]
                factors_batch.weights = torch.as_tensor(factors_batch.weights, device=self.device)
                factors_batch.factors[0] = factors_batch.factors[0][idx_batch]
                model = tl.decomposition.CP_NN_HALS(
                    n_iter_max=self.n_iter_batch,
                    init=factors_batch,
                    **kwargs_tmp,
                )
                model.fit(X_batch)

                ## Drop into self.losses
                # i_total = np.array(list(self.losses.keys()), dtype=int)[:, 1].max() if len(self.losses) > 0 else 0
                # self.losses.update({(i_epoch, i_total + ii): loss.item() for ii, loss in enumerate(model.errors_)})
                self.losses.update({(i_epoch, i_iter): [loss.item() for loss in model.errors_]})


                factors_batch = model.decomposition_

                violation = False
                ## If any of the values are NaN, then throw warning and skip
                if any([torch.any(torch.isnan(f)) for f in factors_batch.factors]) or torch.any(torch.isnan(factors_batch.weights)):
                    print('WARNING: NaNs found. Skipping batch.')
                    violation = True
                ## Jitter any 0s in the factors to be small values
                elif self.jitter_zeros:
                    ## Only do it for the batched factors
                    factors_batch.factors[0] = torch.where(
                        factors_batch.factors[0] <= 0, 
                        1e-6 * torch.rand_like(factors_batch.factors[0]),
                        factors_batch.factors[0],
                    )
                
                if not violation:
                    self.factors.factors[0][idx_batch] = factors_batch.factors[0]
                    self.factors.factors[1:] = factors_batch.factors[1:]
                    self.factors.weights = factors_batch.weights

                if self.verbose:
                    print(f'Epoch {i_epoch+1}/{self.n_epochs}, Batch {i_iter+1}/{len(dataloader)}')
                    print(f'Losses: {[loss.item() for loss in model.errors_]}')
                    print('')
                    
        return self

    def factors_to_numpy(self):
        """
        Returns the factors as numpy arrays.
        """
        self.factors.factors = [f.cpu().numpy() for f in self.factors.factors] if isinstance(self.factors.factors[0], torch.Tensor) else self.factors.factors
        self.factors.weights = self.factors.weights.cpu().numpy() if isinstance(self.factors.weights, torch.Tensor) else self.factors.weights

    @property
    def components_(self):
        """
        Returns the components of the decomposition.
        """
        return self.factors.factors

    def score(self, X):
        """
        Returns the explained variance.
        Reconstructs X_hat using the factors and scores against the input X.

        Args:
            X (torch.Tensor):
                The input data to score against.
        """
        evr = similarity.cp_reconstruction_EVR(
            tensor_dense=X, 
            tensor_CP=self.factors,
        )
        return evr

    def to(self, device):
        """
        Moves the factors to the specified device.

        Args:
            device (str):
                Device to move the factors to.
        """
        if self.factors is not None:
            self.factors.factors = [f.to(device) for f in self.factors.factors]
            self.factors.weights = self.factors.weights.to(device)
        self.device = device
        return self
        

#######################################
########## Incremental PCA ############
#######################################

class IPCA_Dataset(Dataset):
    """
    see incremental_pca for demo
    """
    def __init__(self, 
                 X, 
                 mean_sub=True,
                 zscore=False,
                 preprocess_sample_method='random',
                 preprocess_sample_num=100,
                 device='cpu',
                 dtype=torch.float32):
        """
        Make a basic dataset.
        RH 2021

        Args:
            X (torch.Tensor or np.array):
                Data to make dataset from.
                2-D array. Columns are features, rows are samples.
            mean_sub (bool):
                Whether or not to mean subtract ('center') the
                 columns.
            zscore (bool):
                Whether or not to z-score the columns. This is
                 equivalent to doing PCA on the correlation-matrix.
            preprocess_sample_method (str):
                Method to use for sampling for mean_sub and zscore.
                'random' - uses random samples (rows) from X.
                'first' - uses the first rows of X.
            preprocess_sample_num (int):
                Number of samples to use for mean_sub and zscore.
            device (str):
                Device to use.
            dtype (torch.dtype):
                Data type to use.
        """
        # Upgrade here for using an on the fly dataset.
        self.X = torch.as_tensor(X, dtype=dtype, device=device) # first (0th) dim will be subsampled from
        
        self.n_samples = self.X.shape[0]

        self.mean_sub = mean_sub
        self.zscore = zscore

        if mean_sub or zscore:
            if preprocess_sample_method == 'random':
                self.preprocess_inds = torch.randperm(preprocess_sample_num)[:preprocess_sample_num]
            elif preprocess_sample_method == 'first':
                self.preprocess_inds = torch.arange(preprocess_sample_num)
            else:
                raise ValueError('preprocess_sample_method must be "random" or "first"')
            self.preprocess_inds = self.preprocess_inds.to(device)

        # Upgrade here for using an on the fly dataset.
        if mean_sub:
            self.mean_vals = torch.mean(self.X[self.preprocess_inds,:], dim=0)
        if zscore:
            self.std_vals = torch.std(self.X[self.preprocess_inds,:], dim=0)
        
    def __len__(self):
        return self.n_samples
    
    def __getitem__(self, idx):
        """
        Returns a single sample.

        Args:
            idx (int):
                Index of sample to return.
        """
        if self.mean_sub or self.zscore:
            out = self.X[idx,:] - self.mean_vals
        else:
            out = self.X[idx,:]

        if self.zscore:
            out = out / self.std_vals
            
        return out, idx

def incremental_pca(dataloader,
                    method='sklearn',
                    method_kwargs=None,
                    return_cpu=True,
                    ):
    """
    Incremental PCA using either sklearn or cuml.
    Keep batches small-ish to remain performat (~64 samples).
    RH 2021

    Args:
        dataloader (torch.utils.data.DataLoader):
            Data to be decomposed.
        method (str):
            Method to use.
            'sklearn' : sklearn.decomposition.PCA
            'cuml' : cuml.decomposition.IncrementalPCA
        method_kwargs (dict):
            Keyword arguments to pass to method.
            See method documentation for details.
        device (str):
            Device to use.
            Only used if method is 'cuml'
        return_cpu (bool):  
            Whether or not to force returns/outputs to be on 
             the 'cpu' device. If False, and device!='cpu',
             then returns will be on device.
        return_numpy (bool):
            Whether or not to force returns/outputs to be
             numpy.ndarray type.

    Returns:
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
        EVR (torch.Tensor or np.ndarray):
            The explained variance ratio of each component.
            1-D array.
            Each element is the explained variance ratio of
             the corresponding component.
        object_params (dict):
            Dictionary of parameters used to create the
             decomposition.

    demo:
    import torch_helpers, cupy_helpers

    cupy_helpers.set_device() # calls: cupy.cuda.Device(DEVICE_NUM).use()

    dataset = decomposition.IPCA_Dataset(   X, 
                                            mean_sub=True,
                                            zscore=False,
                                            preprocess_sample_method='random',
                                            preprocess_sample_num=10000,
                                            device='cpu',
                                            dtype=torch.float32)
    dataloader = torch.utils.data.DataLoader(   dataset, 
                                                batch_size=64, 
                                                drop_last=True,
                                                shuffle=False, 
                                                num_workers=0, 
                                                pin_memory=False)

    cuml_kwargs = {
                "handle": cuml.Handle(),
                "n_components": 20,
                "whiten": False,
                "copy": False,
                "batch_size": None,
                "verbose": True,
                "output_type": None
    }

    sk_kwargs = {
                    "n_components": 20,
                    "whiten": False,
                    "copy": False,
                    "batch_size": None,
    }

    components, EVR, ipca = decomposition.incremental_pca(dataloader,
                                    method='cuml',
                                    method_kwargs=cuml_kwargs,
                                    return_cpu=True)
    scores = decomposition.ipca_transform(dataloader, components)
    """
    
    if method_kwargs is None:
        method_kwargs = {}

    if method == 'sklearn':
        ipca = sklearn.decomposition.IncrementalPCA(**method_kwargs)
    elif method == 'cuml':
        ipca =    cuml.decomposition.IncrementalPCA(**method_kwargs)
    
    for iter_batch, batch in enumerate(tqdm(dataloader)):
        if method == 'sklearn':
            batch = batch[0].cpu().numpy()
        if method == 'cuml':
            batch = cupy.asarray(batch[0])
        ipca.partial_fit(batch)
    
    if (return_cpu) and (method=='cuml'):
        components = ipca.components_.get()
    else:
        components = ipca.components_

    EVR = ipca.explained_variance_ratio_

    return components, EVR, ipca

def ipca_transform(dataloader, components):
    """
    Transform data using incremental PCA.
    RH 2020

    Args:
        dataloader (torch.utils.data.DataLoader):
            Data to be decomposed.
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
    """
    out = []
    for iter_batch, batch in enumerate(dataloader):
        out.append(batch[0] @ components.T)
    return torch.cat(out, dim=0)


from typing import Callable, List, Any
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import multiprocessing as mp
from functools import partial
import numpy as np
from tqdm.auto import tqdm

class ParallelExecutionError(Exception):
    """
    Exception class for errors that occur during parallel execution.
    Intended to be used with the ``map_parallel`` function.
    RH 2023

    Attributes:
        index (int):
            Index of the job that failed.
        original_exception (Exception):
            The original exception that was raised.
    """
    def __init__(self, index, original_exception):
        self.index = index
        self.original_exception = original_exception

    def __str__(self):
        return f"Job {self.index} raised an exception: {self.original_exception}"

def map_parallel(
    func: Callable, 
    args: List[Any], 
    method: str = 'multithreading', 
    n_workers: int = -1, 
    n_args: int = None,
    prog_bar: bool = True,
) -> List[Any]:
    """
    Maps a function to a list of arguments in parallel. Includes a progress bar
    and error handling that points to the job that failed.
    RH 2022

    Args:
        func (Callable): 
            The function to be mapped.
        args (List[Any]): 
            List of arguments to which the function should be mapped.
            Length of list should be equal to the number of arguments.
            Each element should then be an iterable for each job that is run.
        method (str): 
            Method to use for parallelization. Either \n
            * ``'multithreading'``: Use multithreading from concurrent.futures.
            * ``'multiprocessing'``: Use multiprocessing from concurrent.futures.
            * ``'mpire'``: Use mpire.
            * ``'pathos'``: Use pathos.multiprocessing.ProcessingPool (similar
            to multiprocessing, but with better serialization support).
            * ``'pytorch'``: Use PyTorch multiprocessing.
            * ``'serial'``: Use list comprehension. \n
            (Default is ``'multithreading'``)
        workers (int): 
            Number of workers to use. If set to -1, all available workers are
            used. (Default is ``-1``)
        n_args (int):
            Number of arguments to expect. If not provided, the length of the
            first element of args is used.
        prog_bar (bool): 
            Whether to display a progress bar using tqdm. (Default is ``True``)

    Returns:
        (List[Any]): 
            output (List[Any]): 
                List of results from mapping the function to the arguments.
                
    Example:
        .. highlight::python
        .. code-block::python

            result = map_parallel(max, [[1,2,3,4],[5,6,7,8]], method='multiprocessing', n_workers=3)
    """
    if n_workers == -1:
        n_workers = mp.cpu_count()

    ## Assert that args is a list
    assert isinstance(args, list), "args must be a list"
    ## Assert that each element of args is an iterable
    assert all([hasattr(arg, '__iter__') for arg in args]), "All elements of args must be iterable"

    ## Get number of arguments. If args is a generator, make None.
    if n_args is None:
        assert hasattr(args[0], '__len__'), f"If n_args is not provided, the first element of args must have a length."
        n_args = len(args[0])
    ## Assert that all args are the same length
    assert all([len(arg) == n_args for arg in args if hasattr(arg, '__len__')]), "All elements of args must have the same length"

    ## Make indices
    indices = np.arange(n_args)

    ## Prepare args_map (input to map function)
    args_map = [[func] * n_args, *args, indices]
        
    if method == 'multithreading':
        executor = ThreadPoolExecutor
        star_args = True
    elif method == 'multiprocessing':
        executor = ProcessPoolExecutor
        star_args = True
    elif method == 'mpire':
        import mpire
        executor = mpire.WorkerPool
        star_args = False
    elif method == 'pathos':
        from pathos.multiprocessing import ProcessingPool
        executor = ProcessingPool
        star_args = True
    elif method == 'pytorch':
        from torch.multiprocessing import Pool
        executor = Pool
        star_args = False
    elif method == 'serial':
        return    list(tqdm(map(_func_wrapper_helper, *args_map), total=n_args, disable=prog_bar!=True))
    else:
        raise ValueError(f"method {method} not recognized")

    with executor(n_workers) as ex:
        if star_args:
            return list(tqdm(ex.map(_func_wrapper_helper, *args_map), total=n_args, disable=prog_bar!=True))
        else:
            return list(tqdm(ex.map(_func_wrapper_helper, args_map), total=n_args, disable=prog_bar!=True))
def _func_wrapper_helper(*func_args_index):
    """
    Wrapper function to catch exceptions.
    
    Args:
    *func_args_index (tuple):
        Tuple of arguments to be passed to the function.
        Should take the form of (func, arg1, arg2, ..., argN, index)
        The last element is the index of the job.
    """
    func = func_args_index[0]
    args = func_args_index[1:-1]
    index = func_args_index[-1]

    try:
        return func(*args)
    except Exception as e:
        raise ParallelExecutionError(index, e)
    
    

def multiprocessing_pool_along_axis(x_in, function, n_workers=None, axis=0, **kwargs):
    pool = mp.Pool(processes=n_workers)
    if axis==0:
        results = pool.map(partial(function , **kwargs), [x_in[ii] for ii in range(x_in.shape[0])])
        pool.close()
        pool.join()
        return np.row_stack(results)
    elif axis==1:
        results = pool.map(partial(function , **kwargs), [x_in[:,ii] for ii in range(x_in.shape[1])])
        pool.close()
        pool.join()
        return np.column_stack(results)

def unpacking_apply_along_axis(all_args):
    """
    Like numpy.apply_along_axis(), but with arguments in a tuple
    instead.

    This function is useful with multiprocessing.Pool().map(): (1)
    map() only handles functions that take a single argument, and (2)
    this function can generally be imported from a module, as required
    by map().
    """
    (func1d, axis, arr, args, kwargs) = all_args
    return np.apply_along_axis(func1d, axis, arr, *args, **kwargs)
def parallel_apply_along_axis(func1d, axis, arr, *args, **kwargs):
    """
    Like numpy.apply_along_axis(), but takes advantage of multiple
    cores.
    Taken from here: https://stackoverflow.com/questions/45526700/easy-parallelization-of-numpy-apply-along-axis
    """        
    # Effective axis where apply_along_axis() will be applied by each
    # worker (any non-zero axis number would work, so as to allow the use
    # of `np.array_split()`, which is only done on axis 0):
    effective_axis = 1 if axis == 0 else axis
    if effective_axis != axis:
        arr = arr.swapaxes(axis, effective_axis)

    # Chunks for the mapping (only a few chunks):
    chunks = [(func1d, effective_axis, sub_arr, args, kwargs)
              for sub_arr in np.array_split(arr, mp.cpu_count())]

    pool = mp.Pool()
    individual_results = pool.map(unpacking_apply_along_axis, chunks)
    # Freeing the workers:
    pool.close()
    pool.join()

    return np.concatenate(individual_results)



from typing import Dict, Type, Any, Union, Optional, Callable, Tuple, List
import gc

import numpy as np
import sklearn
import sklearn.linear_model


def LinearRegression_sweep(X_in,
                            y_in,
                            cv_idx,
                            alphas=0,
                            l1_ratios=0.5,
                            rolls=0,
                            method_package='sklearn',
                            method_model='LinearRegression',
                            compute_preds=False,
                            verbose=True,
                            theta_inPlace=None, 
                            intercept_inPlace=None, 
                            EV_train_inPlace=None, 
                            EV_test_inPlace=None,
                            preds_inPlace=None,
                            **model_params,
                            ):
    '''
    Performs linear regression over a sweep of input parameters.
    For every input parameter, each entry linearly increases the
    size of the parameter grid. The number of regression runs and
    shape of the outputs will be:
        n_y * n_alphas * n_l1_ratios * n_cv_idx * n_rolls.

    This function is similar to using sklearn's CV-GridSearch
    functions. Benefits to this function are that you can use
    a variety of different models, including CuML's functions.
    Also allows for in-place computation so you can stop it early
    and tune the model easier.
    RH 2021

    Args:
        X (ndarray): 
            Predictors. Shape: (n_samples, n_features).
        y (ndarray): 
            Outputs. Regression performed on one column at a time.
            Shape: (n_samples, n_outputs).
        alphas (ndarray): 
            Alpha parameters. Sets strength of regularization.
        l1_ratios (ndarray): 
            L1 to L2 ratio parameters. 0:L2 only, 1:L1 only
        cv_idx (list of lists of 1-D arrays): 
            Cross-validation indices. Each entry in the outer list is
            for a different run. Each entry in the outer list should 
            contain 2 lists, the first one containing the training 
            indices, and the second one containing the test indices.
            The function 'cross_validation.make_cv_indices' can make
            this variable easily.
        rolls (ndarray of ints):
            1-D array containing index shifts to apply.
        method_package (string):
            Can be any of the following:
                'cuml': 
                    Uses Rapids' cuml.linear_model library (requires GPU and CuPy)
                'sklearn':
                    Uses sklearn's sklearn.linear_model library
        method_model (string):
            Can be any of the following functions:
                'LinearRegression', 'Ridge', 'Lasso', 'ElasticNet'
            Note that not all models use the same parameters (eg LinearRegression
            does not use l1_ratio or alpha). So, if there are multiple
            values for an unused parameter, then the regression will be
            unnecessarily computed multiple times.
        compute_preds (bool):
            If True, will compute the predictions for the model.
             preds = theta @ X.T + intercept
            This is currently done on CPU so it is very slow.
        verbose (int 0-2):
            Preference of whether to print reconstruction scores.
            0: Print nothing
            1: Print for every y input
            2: Print at every regression step
        theta_inPlace (ndarray): (optional)
            Allows for stopping the function while it is running and 
            recovering the data up to that point. 
            Input a pre-allocated array with appropriate shape (see function below)
        intercept_inPlace (ndarray): (optional)
            Allows for stopping the function while it is running and 
            recovering the data up to that point. 
            Input a pre-allocated array with appropriate shape (see function below)
        EV_train_inPlace (ndarray): (optional)
            Allows for stopping the function while it is running and 
            recovering the data up to that point. 
            Input a pre-allocated array with appropriate shape (see function below)
        EV_test_inPlace (ndarray): (optional)
            Allows for stopping the function while it is running and 
            recovering the data up to that point. 
            Input a pre-allocated array with appropriate shape (see function below)
        **model_params (**kwargs):
            Pass into this function as **params.
            Different models require different parameters beyond alphas and 
            l1_ratios. You should input those parameters here.
            Check documentation for sklearn/CuML's documentation for what to
            put in here.

    Returns:
        theta (ndarray):
            Regression coefficients.
            Shape: (n_y, n_splits, n_rolls, n_alphas, n_l1Ratios, X.shape[1])
        intercept (ndarray):
            Intercept values
            Shape: (n_y, n_splits, n_rolls, n_alphas, n_l1Ratios)
        EV_test (ndarray):
            Pearson correlation scores between test data and reconstructed
            test data.
            Shape: (n_y, n_splits, n_rolls, n_alphas, n_l1Ratios)
        EV_train (ndarray):
            Pearson correlation scores between train data and reconstructed
            train data.
            Shape: (n_y, n_splits, n_rolls, n_alphas, n_l1Ratios)

    ============ DEMO ===============

    from basic_neural_processing_modules import cross_validation

    from sklearn.model_selection import (TimeSeriesSplit, KFold, ShuffleSplit,
                                    StratifiedKFold, GroupShuffleSplit,
                                    GroupKFold, StratifiedShuffleSplit)
    group_len = 60*2 * Fs # seconds * Fs
    n_splits = 10
    test_size = 0.3
    groups = np.arange(X.shape[0])//group_len
    n_groups = np.max(groups)
    cv = GroupShuffleSplit(n_splits, test_size=test_size)
    cv_idx = cross_validation.make_cv_indices(cv,
                                            groups,
                                            lw=10,
                                            plot_pref=True)

    OR:

    n_splits = 4
    cv = StratifiedKFold(n_splits, shuffle=False, random_state=None, )
    cv_idx = list(cv.split(X=scores.T, y=trial_types_aligned, groups=trial_types_aligned))


    from basic_neural_processing_modules.linear_regression import LinearRegression_sweep

    n_nonzero_rolls = 2
    min_roll = 60*10*Fs
    max_roll = X.shape[0] - min_roll
    rolls = np.concatenate(([0] , np.random.randint(min_roll, max_roll, n_nonzero_rolls)))
    n_rolls = n_nonzero_rolls + 1

    model_params_cuml_ElasticNet = {
            'fit_intercept': True,
            'normalize': False,
            'max_iter': 1000,
            'tol': 0.0001,
            'selection': 'cyclic',
    }

    l1_ratios = np.array([0, 0.01, 0.1, 0.5, 0.9, 0.99, 1])
    alphas = np.array([0.0001, 0.001, 0.01, 0.1, 1])

    # prepare output variables for in-place computations
    n_y = y.shape[1]
    n_alphas = len(alphas)
    n_l1Ratios = len(l1_ratios)

    theta = np.ones((n_y , n_splits , n_rolls , n_alphas , n_l1Ratios, X.shape[1]))
    intercept = np.ones((n_y , n_splits , n_rolls , n_alphas , n_l1Ratios))
    EV_train  = np.zeros((n_y, n_splits , n_rolls , n_alphas , n_l1Ratios))
    EV_test   = np.zeros((n_y, n_splits , n_rolls , n_alphas , n_l1Ratios))

    # Run regression sweep
    theta, intercept, EV_train, EV_test, preds = LinearRegression_sweep(   X,
                                                                    y,
                                                                    cv_idx,
                                                                    alphas=alphas,
                                                                    l1_ratios=l1_ratios,
                                                                    rolls=rolls,
                                                                    method_package='cuml',
                                                                    method_model='ElasticNet',
                                                                    compute_preds=True,
                                                                    verbose=True,
                                                                    theta_inPlace=theta, 
                                                                    intercept_inPlace=intercept,
                                                                    EV_train_inPlace=EV_train,
                                                                    EV_test_inPlace=EV_test,
                                                                    **model_params_cuml_ElasticNet
                                                                    )
'''

    gc.collect()
    
    if method_package in ['cuml']:
        import cudf
        import cuml
        import cupy

        print('making cupy arrays')    
        X = cupy.asarray(X_in)
        y = cupy.asarray(y_in)

        # I thought this might help with the memory leak issue,
        # but it doesn't seem to help that much
        model_params['handle'] = cuml.Handle()
    else:
        X = np.array(X_in)
        y = np.array(y_in)
    

    n_alphas = len(alphas)
    n_l1Ratios = len(l1_ratios)

    n_splits = len(cv_idx)
    len_train = len(cv_idx[0][0])
    len_test = len(cv_idx[0][1])

    n_rolls = len(rolls)

    n_y = y.shape[1]

    if method_model in ['LogisticRegression']:
        n_classes = len(np.unique(y))
    else:
        n_classes = 1


    if theta_inPlace is not None:
        theta = theta_inPlace
    else:
        theta = np.ones((n_y , n_splits , n_rolls , n_alphas , n_l1Ratios, n_classes, X.shape[1]))
    
    if preds_inPlace is not None:
        preds = preds_inPlace
    else:
        preds = np.zeros((n_y, n_splits , n_rolls , n_alphas , n_l1Ratios, n_classes, X.shape[0]))


    if intercept_inPlace is not None:
        intercept = intercept_inPlace
    else:
        intercept = np.ones((n_y , n_splits , n_rolls , n_alphas , n_l1Ratios))
    
    if EV_train_inPlace is not None:
        EV_train = EV_train_inPlace
    else:
        EV_train  = np.zeros((n_y, n_splits , n_rolls , n_alphas , n_l1Ratios))
    
    if EV_test_inPlace is not None:
        EV_test = EV_test_inPlace
    else:
        EV_test   = np.zeros((n_y, n_splits , n_rolls , n_alphas , n_l1Ratios))

    for iter_factor in range(n_y):
        for iter_roll in range(n_rolls):
            y_iter = np.roll(y[:,iter_factor] , rolls[iter_roll])
            
            for iter_cv, (idx_train , idx_test) in enumerate(cv_idx):
                X_train = X[idx_train,:]
                y_train = y_iter[idx_train]
                X_test = X[idx_test,:]
                y_test = y_iter[idx_test]

                for iter_alpha , alpha in enumerate(alphas):
                    for iter_l1Ratio , l1_ratio in enumerate(l1_ratios):
                        if method_model in ['LinearRegression']:
                            clf = eval(f'{method_package}.linear_model.{method_model}')(**model_params)
                        
                        if method_model in  ['Lasso', 'Ridge']:
                            clf = eval(f'{method_package}.linear_model.{method_model}')(alpha=alpha,
                                                                                        **model_params)
                        
                        if method_model in ['ElasticNet']:
                            clf = eval(f'{method_package}.linear_model.{method_model}')(alpha=alpha,
                                                                                        l1_ratio=l1_ratio,
                                                                                        **model_params)

                        if method_model in ['LogisticRegression']:
                            clf = eval(f'{method_package}.linear_model.{method_model}')(C=1/alpha,
                                                                                        l1_ratio=l1_ratio if model_params['penalty'] == 'elasticnet' else None,
                                                                                        **model_params)

                        clf.fit(X_train , y_train )

                        if method_package=='cuml':
                            # theta[:, iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = cupy.asnumpy(clf.coef_)
                            # intercept[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = cupy.asnumpy(clf.intercept_)
                            theta[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = cupy.asnumpy(clf.coef_)
                            intercept[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = cupy.asnumpy(clf.intercept_)
                        else:
                            # theta[:, iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = clf.coef_
                            # intercept[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = clf.intercept_
                            theta[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = clf.coef_
                            intercept[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = clf.intercept_

                        EV_train_tmp = clf.score(X_train, y_train)
                        EV_test_tmp = clf.score(X_test, y_test)
                        
                        EV_train[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = EV_train_tmp
                        EV_test[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = EV_test_tmp       
                        
                        if compute_preds:
                            if method_package=='cuml':
                                preds[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = ((clf.coef_ @ X.T) + clf.intercept_).get()
                            else:
                                preds[iter_factor, iter_cv, iter_roll, iter_alpha, iter_l1Ratio] = ((clf.coef_ @ X.T) + clf.intercept_)
                        else:
                            preds = np.nan

                        if verbose==2:
                            print(f'y #: {iter_factor} , Roll iter: {iter_roll} , CV repeat #: {iter_cv} , alpha val: {alpha} , l1_ratio: {l1_ratio} , train R^2: {round(EV_train_tmp,3)}')
                            print(f'y #: {iter_factor} , Roll iter: {iter_roll} , CV repeat #: {iter_cv} , alpha val: {alpha} , l1_ratio: {l1_ratio} , test  R^2: {round(EV_test_tmp,3)} \n')
        if verbose==1:
            print(f'computed y #: {iter_factor}')

            # This is pretty important when using CuML on GPU. 
            # Move this line of code up or down within the for-loops to make it more or less frequent.
            # Generally move it as far up the for-loop hierarchy as possible without running out of memory
            gc.collect()

    return theta, intercept, EV_train, EV_test, preds


import numpy as np
from tqdm import trange

#################################################################################
######################### START OF HELPER FUNCTIONS #############################
#################################################################################


def update_varSum(idx_new, series, win_size, prevVal=None):
    """
    Returns the power sum average based on the blog post from
    Subliminal Messages.  Use the power sum average to help derive the running
    variance.
    sources: http://subluminal.wordpress.com/2008/07/31/running-standard-deviations/

    Keyword arguments:
    idx_new     --  current index or location of the value in the series
    series  --  list or tuple of data to average
    win_size  -- number of values to include in average
    prevVal    --  previous powersumavg (n - 1) of the series.
    """

    if win_size < 1:
        raise ValueError("period must be 1 or greater")

    if idx_new < 0:
        idx_new = 0

    if np.any(prevVal == None):
        if idx_new > 0:
            raise ValueError("pval of None invalid when bar > 0")

        prevVal = np.zeros(series.shape[1])

    newVal = np.double(series[idx_new,:])

    if idx_new < win_size:
        result = prevVal + (newVal * newVal - prevVal) / (idx_new + 1.0)

    else:
        oldVal = np.double(series[idx_new - win_size,:])
        result = prevVal + (((newVal * newVal) - (oldVal * oldVal)) / win_size)

    return result

def varSum_to_var(idx_new, series, win_size, mean_current, varSum):
    """
    Returns the running variance based on a given time period.
    sources: http://subluminal.wordpress.com/2008/07/31/running-standard-deviations/

    Keyword arguments:
    idx_new     --  current index or location of the value in the series
    series  --  list or tuple of data to average
    mean_current    --  current average of the given period
    varSum -- current powersumavg of the given period
    """
    if win_size < 1:
        raise ValueError("period must be 1 or greater")

    if idx_new <= 0:
        return 0.0

    if np.any(mean_current == None):
        raise ValueError("asma of None invalid when bar > 0")

    if np.any(varSum == None):
        raise ValueError("powsumavg of None invalid when bar > 0")

    windowsize = idx_new + 1.0
    if windowsize >= win_size:
        windowsize = win_size

    return (varSum * windowsize - windowsize * mean_current * mean_current) / windowsize

def running_mean(idx_new, series, mean_old):
    """
    Returns the cumulative or unweighted simple moving average.
    Avoids sum of series per call.

    Keyword arguments:
    idx_new     --  current index or location of the value in the series
    series  --  list or tuple of data to average
    mean_old  --  previous average (n - 1) of the series.
    """

    if idx_new <= 0:
        return series[0,:]

    return mean_old + ((series[idx_new,:] - mean_old) / (idx_new + 1.0))

def update_mean(idx_new, series, win_size, mean_old):
    """
    Returns the running simple moving average - avoids sum of series per call.

    Keyword arguments:
    idx_new     --  current index or location of the value in the series
    series  --  list or tuple of data to average
    win_size  --  number of values to include in average
    mean_old  --  previous simple moving average (n - 1) of the series
    """

    if win_size < 1:
        raise ValueError("period must be 1 or greater")

    if idx_new <= 0:
        return series[0,:]

    elif idx_new < win_size:
        return running_mean(idx_new, series, mean_old)

    return mean_old + ((series[idx_new,:] - series[idx_new - win_size,:]) / float(win_size))


#################################################################################
######################### END OF HELPER FUNCTIONS ###############################
###################### START OF PROCESSING FUNCTIONS ############################
#################################################################################


def make_rollingZScore(X , win_roll):

    X_mean_rolling = np.ones_like(X) * np.nan
    X_var_rolling = np.ones_like(X) * np.nan
    win_size_rollingBaseline = win_roll
    list_of_values = X
    varSum_old = None
    mean_old = None
    for idx in trange(len(list_of_values)):

        mean_new = update_mean(idx, list_of_values, win_size_rollingBaseline, mean_old)
        varSum_new = update_varSum(idx, list_of_values, win_size_rollingBaseline, varSum_old)
        var_new = varSum_to_var(idx, list_of_values, win_size_rollingBaseline, mean_new, varSum_new)

        X_mean_rolling[idx,:] = mean_new
        X_var_rolling[idx,:] = var_new

        mean_old = mean_new
        varSum_old = varSum_new

    eps = 1e-7
    X_mean_rolling[X_mean_rolling<eps] = eps
    X_zscore_roll = (list_of_values - X_mean_rolling)/np.sqrt(X_var_rolling)

    return X_zscore_roll